# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import collections
import json
import logging
import random
import uuid

from majormode.perseus.constant.account import AccountType
from majormode.perseus.constant.obj import ObjectStatus
from majormode.perseus.constant.contact import ContactName
from majormode.perseus.model import obj
from majormode.perseus.model.contact import Contact
from majormode.perseus.model.locale import DEFAULT_LOCALE
from majormode.perseus.model.locale import Locale
from majormode.perseus.model.obj import Object
from majormode.perseus.utils import module_util
from majormode.perseus.utils.date_util import ISO8601DateTime
from majormode.perseus.utils.rdbms import RdbmsConnection

from majormode.perseus.service.base_rdbms_service import BaseRdbmsService
from majormode.perseus.service.base_service import BaseService
from majormode.perseus.utils import cast
from majormode.perseus.utils import rdbms


class ContactService(BaseRdbmsService):
    class ContactAlreadyInUseException(BaseService.BaseServiceException):
        """
        Signal that the specified contact information is already associated
        with an another user account of the platform.
        """
        pass

    class UnverifiedContactException(BaseService.BaseServiceException):
        """
        Signal that the contact information provided by the user has not been
        verified yet and it CANNOT be used for the ongoing procedure.
        """
        pass

    # Define the minimal allowed duration of time expressed in seconds
    # between two consecutive requests to verify a same contact information.
    MINIMAL_TIME_BETWEEN_CONTACT_VERIFICATION_REQUEST = 60 * 5

    # Maximum number of digits that composed a verification code used to
    # validate a contact information.
    VERIFICATION_CODE_MAXIMUM_DIGIT_COUNT = 9

    def __cleanse_contact_references(
            self,
            account_id,
            contact,
            connection=None):
        """
        Remove the specified contact information from any other user account
        but the one passed to this function.

        The function also cancels any verification requests  of this contact
        information that would be still ongoing.


        :param account_id: identification of the account of the user who has
            confirmed the given contact information.

        :param contact: an instance `Contact` of the contact information that
            the user has confirmed, and which references to it need to be
            removed from any other user account and ongoing verification
            request.

        :param connection: An object `RdbmsConnection`  with auto commit.
        """
        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            # Remove this contact information from any other user account.
            cursor = connection.execute("""
                DELETE FROM 
                    account_contact
                  WHERE
                    property_name = %(property_name)s
                    AND property_value = %(property_value)s
                    AND account_id <> %(account_id)s
                  RETURNING
                    account_id""",
                {
                    'account_id': account_id,
                    'property_name': contact.property_name,
                    'property_value': contact.property_value
                })

            account_ids = [
                row.get_value('account_id', cast.string_to_uuid)
                for row in cursor.fetch_all()
            ]

            # Soft-delete accounts that this contact information has been removed
            # from and that have no contact information anymore.
            if account_ids:
                connection.execute(
                    """
                    UPDATE
                        account
                      SET 
                        object_status = %(OBJECT_STATUS_DISABLED)s
                      WHERE
                        account_id IN (%(account_ids)s)
                        AND NOT EXISTS(
                          SELECT
                              true
                            FROM
                              account_contact
                            WHERE
                             account_contact.account_id = account.account_id)
                    """,
                    {
                        'OBJECT_STATUS_DISABLED': ObjectStatus.disabled,
                        'account_ids': account_ids,
                    })

            # Delete any other verification requests of this contact information.
            connection.execute(
                """
                DELETE FROM 
                    account_contact_verification
                  WHERE
                    property_name = %(property_name)s
                    AND property_value = %(property_value)s""",
                {
                    'property_name': contact.property_name,
                    'property_value': contact.property_value,
                })

    def __enable_account_contact(
            self,
            account_id: uuid.UUID,
            contact: Contact,
            connection: rdbms.RdbmsConnection = None):
        """
        Enable the contact information of the specified user


        :param account_id: The identification of the account of a user.

        :param contact: The contact information to be enabled.

        :param connection: An object `RdbmsConnection` with auto commit.

        """
        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            cursor = connection.execute(
                """
                UPDATE 
                    account_contact
                  SET 
                    is_verified = true,
                    update_time = current_timestamp
                  WHERE
                    account_id = %(account_id)s
                    AND property_name = %(property_name)s
                    AND property_value = %(property_value)s
                  RETURNING
                    is_primary
                  """,
                {
                    'account_id': account_id,
                    'property_name': contact.property_name,
                    'property_value': contact.property_value
                })

            is_primary = cursor.fetch_one().get_value('is_primary')

            # Set this contact information as primary if the user doesn't have any
            # primary contact information.
            if not is_primary:
                cursor = connection.execute(
                    """
                    SELECT EXISTS(
                      SELECT
                          true
                        FROM
                          account_contact
                        WHERE
                          account_id = %(account_id)s
                          AND property_name = %(property_name)s
                          AND is_primary)
                    """,
                    {
                        'account_id': account_id,
                        'property_name': contact.property_name
                    })

                if cursor.get_row_count() == 0:
                    self.set_primary_contact(account_id, contact, connection=connection)

    @staticmethod
    def __generate_nonce(digit_count):
        """
        Generate a string composed of the requested number of digits


        :param digit_count: Number of digit to generate.


        :return: A string composed of the request number of digits.
        """
        return ''.join([
            str(random.randint(0, 9))
            for _ in range(digit_count)
        ])

    def __get_contact_verification_request(
            self,
            connection: rdbms.RdbmsConnection = None,
            contact: Contact = None,
            verification_code: str = None,
            request_id: uuid.UUID = None):
        """

        :param connection:
        :param contact:
        :param verification_code:
        :param request_id: The identification of the verification request.


        :return:


        :raise DisabledObjectException: If the verification request has
            expired.

        :raise UndefinedObjectException: If the verification request doesn't
            exist.
        """
        with self.acquire_rdbms_connection(connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    action_type,
                    attempt_count,
                    context,
                    creation_time,
                    expiration_time,
                    last_attempt_time,
                    object_status,
                    request_count,
                    request_id,
                    update_time
                  FROM
                    account_contact_verification
                  WHERE
                    (%(request_id)s IS NOT NULL AND request_id = %(request_id)s)
                    OR (%(verification_code)s IS NOT NULL 
                        AND verification_code = %(verification_code)s
                        AND property_name = %(property_name)s
                        AND property_value = %(property_value)s)
                """,
                {
                    'property_name': contact.property_name,
                    'property_value': contact.property_value,
                    'request_id': request_id,
                    'verification_code': verification_code,
                })

            row = cursor.fetch_one()
            request = row and row.get_object({
                'creation_time': cast.string_to_timestamp,
                'expiration_time': cast.string_to_timestamp,
                'last_attempt_time': cast.string_to_timestamp,
                'object_status': ObjectStatus,
                'request_id': cast.string_to_uuid,
                'update_time': cast.string_to_timestamp,
            })

            if request is None or request.object_status == ObjectStatus.deleted:
                raise self.UndefinedObjectException(f"The verification request doesn't exist")

            # Delete this request if it has expired.
            if request.expiration_time and ISO8601DateTime.now() >= request.expiration_time:
                self.__update_contact_verification_request(request.request_id, ObjectStatus.disabled)
                request.object_status = ObjectStatus.disabled

            if request.object_status == ObjectStatus.disabled:
                raise self.DisabledObjectException(f"The verification request has expired")

            return request

    def __update_contact_verification_request(
            self,
            request_id,
            connection=None,
            object_status=None):
        """
        Update a contact verification request status


        :param request_id: Identification of a contact verification request.

        :param connection: An object `RdbmsConnection` with auto commit.

        :param object_status: An item of the enumeration `ObjectStatus`

            * `deleted`:  The request has been fulfilled.

            * `disabled`: The request has expired.
        """
        if object_status not in (ObjectStatus.deleted, ObjectStatus.disabled):
            raise ValueError("Wrong value of the argument 'object_status'")

        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            connection.execute(
                """
                UPDATE 
                    account_contact_verification
                  SET
                    object_status = %(object_status)s,
                    update_time = current_timestamp
                  WHERE
                    request_id = %(request_id)s
                """,
                {
                    'object_status': object_status,
                    'request_id': request_id
                })

    def is_contact_available(
            self,
            contact,
            account_id=None,
            connection=None):
        """
        Indicate whether the specified contact information is available and
        verified.

        In the case the contact information would have been already registered
        by a user, indicate whether this contact information has been verified.


        :param contact: An object `Contact`.

        :param account_id: Identification of the account of the user on behalf
            whom this function is called.

        :param connection: An instance `RdbmsConnection` to be used
            supporting the Python clause `with ...:`.


        :return: A tuple containing the following values:

            * `is_available`: `True` if the specified contact information is not
              yet registered by any accounts, or if it has been registered by the
              account on behalf of whom this function is called; `False` otherwise.

            * `is_verified`: `True` if the specified contact information has been
               verified; `False` otherwise.
        """
        with self.acquire_rdbms_connection(connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    is_verified,
                    account_id
                  FROM 
                    account_contact
                  WHERE 
                    property_name = %(property_name)s
                    AND property_value = %(property_value)s
                    AND object_status = %(OBJECT_STATUS_ENABLED)s
                """,
                {
                    'OBJECT_STATUS_ENABLED': ObjectStatus.enabled,
                    'property_name': contact.property_name,
                    'property_value': contact.property_value,
                })
            row = cursor.fetch_one()

            # Check whether this contact information is associated to any accounts.
            if row is None:
                return True, False

            # Retrieve the verification status of this contact information and the
            # account identification of the user who has registered it .
            contact = row.get_object({'account_id': cast.string_to_uuid})

            # If `account_id` has been passed to this function, check whether this
            # user has registered this contact information, in which case this
            # contact information is declared available even if already verified.
            return False if account_id is None else account_id == contact.account_id, contact.is_verified

    def request_contact_change(
            self,
            app_id: uuid.UUID,
            account_id: uuid.UUID,
            old_contact: Contact,
            new_contact: Contact,
            connection: RdbmsConnection = None,
            context: dict | list | set = None,
            language: Locale = DEFAULT_LOCALE,
            verification_code_length: int = None):
        if old_contact.property_name != new_contact.property_value:
            logging.error(
                f"Incompatible old contact information {old_contact.property_name} "
                f"with new contact information {new_contact.property_name}"
            )
            raise self.InvalidOperationException(f"Type mismatch between the two contacts")

        if not self.has_contact(account_id, old_contact, connection=connection):
            logging.error(f"The account {account_id} is not associated with the contact {old_contact.property_value}")
            raise self.IllegalAccessException(f"The user is not associated with the specified contact")

        if not self.is_contact_available(new_contact, connection=connection):
            logging.error(f"The contact {new_contact.property_value} is already in use")
            raise self.InvalidOperationException("The new requested contact is already in use")

        verification_code = self.__generate_nonce(min(
            verification_code_length,
            self.VERIFICATION_CODE_MAXIMUM_DIGIT_COUNT
        ))

        expiration_time = None

        cursor = connection.execute(
            """
            INSERT INTO account_contact_change_request (
                    account_id,
                    app_id,
                    context,
                    expiration_time,
                    language,
                    property_name,
                    property_new_value,
                    property_old_value,
                    verification_code
                )
                VALUES (
                )
                RETURNING
                  request_id,
                  update_time
            """,
            {
                'account_id': account_id,
                'app_id': app_id,
                'context': context,
                'expiration_time': expiration_time,
                'language': language,
                'property_name': old_contact.property_name,
                'property_new_value': new_contact.property_value,
                'property_old_value': old_contact.property_value,
                'verification_code': verification_code,
            }
        )


    def request_contact_verification(
            self,
            contact: Contact,
            account_id: uuid.UUID = None,
            connection: RdbmsConnection = None,
            context: dict | list | set = None,
            language: Locale = DEFAULT_LOCALE,
            verification_code_length: int = None):
        """
        Request the initiation of the process to verify the specified contact
        information.

        A background task running on the server platform is responsible for
        sending the verification request to the contact address the user has
        provided.  This message contains more likely a HTML link that will
        redirect the user to a web page responsible for confirming this
        contact information (cf. function `confirm_contact`).


        :note: The function ignores any consecutive call for a same contact
            information within a minimal duration of time, to avoid spamming
            the user.

        :note: If a verification request was already generated on behalf of
            another user, the function will silently disable the association
            between this request and this particular user account.  The client
            application MUST request the user who will confirm this contact
            information will have to authenticate in order to add this contact
            information to his account.


        :warning: The function DOESN'T return the identification of the
            verification request so that this identification CANNOT be passed
            to the client application.  Indeed!


        :param contact: An instance `Contact`.

        :param account_id: Identification of the account of the user who
            requests to get his contact information verified.  This argument
            can be `None` when this contact information needs to be verified
            prior to the user account to be created.

        :param connection: A `RdbmsConnection` instance to be used  supporting
            the Python clause `with ...:`.

        :param context: A JSON expression corresponding to the context in
            this contact information has been added and needs to be verified.

        :param language: An object `Locale` referencing the language in which
            to generate the contact information verification request.

        :param verification_code_length: Number of digits to generate the
            verification code.


        :raise IllegalAccessException: If a verification of this contact
            information has been already requested for another account.
        """
        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            # Check whether a verification request for this contact information is
            # ongoing.
            cursor = connection.execute(
                """
                SELECT 
                    request_id,
                    account_id,
                    EXTRACT(EPOCH FROM current_timestamp - update_time) AS elapsed_time
                  FROM 
                    account_contact_verification
                  WHERE
                    property_name = %(property_name)s
                    AND property_value = %(property_value)s
                """,
                {
                    'property_name': contact.property_name,
                    'property_value': contact.property_value
                })

            row = cursor.fetch_one()
            request = row and row.get_object({
                'account_id': cast.string_to_uuid,
                'request_id': cast.string_to_uuid
            })

            # Generate a verification request for this contact information if none
            # has been generated so far.
            if request is None:
                verification_code = self.__generate_nonce(
                    min(verification_code_length, self.VERIFICATION_CODE_MAXIMUM_DIGIT_COUNT))

                connection.execute(
                    """
                    INSERT INTO account_contact_verification(
                        account_id,
                        context,
                        language,
                        property_name,
                        property_value,
                        verification_code
                      )
                      VALUES (
                        %(account_id)s,
                        %(context)s,
                        %(language)s,
                        %(property_name)s,
                        %(property_value)s,
                        %(verification_code)s
                      )
                    """,
                    {
                        'account_id': account_id,
                        'context': json.dumps(obj.stringify(context, trimmable=True)),
                        'language': language,
                        'property_name': contact.property_name,
                        'property_value': contact.property_value,
                        'verification_code': verification_code,
                    })

            else:
                # If a verification request was already generated on behalf of another
                # user, we disable the association between this request and a
                # particular user account.  The user who will confirm this contact
                # information will have to authenticate to add this contact information
                # to his account.
                if request.account_id and account_id and request.account_id != account_id:
                    connection.execute(
                        """
                        UPDATE 
                            account_contact_verification
                          SET
                            account_id = NULL,
                            update_time = current_timestamp
                          WHERE
                            request_id = %(request_id)s
                        """,
                        {
                            'request_id': request.request_id
                        })

                # Ignore successive calls for sending a new message to the user that
                # would be requested with a minimal duration of time, to avoid
                # spamming the user.
                if request.elapsed_time > self.MINIMAL_TIME_BETWEEN_CONTACT_VERIFICATION_REQUEST:
                    connection.execute(
                        """
                        UPDATE 
                            account_contact_verification
                          SET
                            request_count = request_count + 1,
                            update_time = current_timestamp
                          WHERE
                            request_id = %(request_id)s
                          """,
                        {
                            'request_id': request.request_id
                        })

    def validate_contact_verification_request(
            self,
            connection: rdbms.RdbmsConnection = None,
            contact: Contact = None,
            verification_code: str = None,
            request_id: uuid.UUID = None):
        """
        Indicate whether the specified contact verification request exists.


        :param connection: An object `RdbmsConnection`.

        :param contact: The contact information that the user is verifying.
            The argument `activation_code` MUST be also passed to this
            function.

        :param request_id: identification of the contact verification request.

        :param verification_code: A pseudo-random number (nonce) that was
            generated and sent to the user to verify his contact information.
            The argument `contact` MUST be also passed to this function.


        :return: `True` if the specified identification corresponds to a
            contact verification request registered to the platform;
            `False` otherwise.


        :raise DisabledObjectException: If the contact verification request
            has expired.

        :raise IllegalAccessException: If the verification code is not valid.

        :raise UndefinedObjectException: If the verification code doesn't
            exist.  It may be an expired request that has been deleted.
        """
        request = self.__get_contact_verification_request(
            contact=contact,
            verification_code=verification_code)

        return request

















    @staticmethod
    def __group_contacts_by_property_name(contacts):
        """
        Return a dictionary of contacts group by their name.


        :param contacts: A list of objects `Contact`.


        :return: A dictionary where the key corresponds to an item of
            `ContactName` and the value a list of contacts that correspond
            to this contact type.


        :raise AssertError: If a contact information is duplicated.
        """
        contacts_by_property_name = collections.defaultdict(list)

        for contact in contacts:
            assert contact.property_value not in [
                _contact.property_value
                for _contact in contacts_by_property_name[contact.property_name]
            ], \
                'Duplicated contact information'
            contacts_by_property_name[contact.property_name].append(contact)

        return contacts_by_property_name

    @classmethod
    def __subtract_contacts(cls, from_contacts, with_contacts):
        """
        Return a new list of contacts in the first list that are not in the
        second list.


        :note: the function also automatically sets the attribute `is_primary`
            of contact instances to `False` when a contact of the same name is
            in the second list and its attribute `is_primary` is `True`.


        :param from_contacts: a list of contacts to subtract contacts that are
            in the second list.

        :param with_contacts: a list of contacts to be removed from the first
            list of contacts.


        :return: a list with contacts in the first list that are not in the
            second list.
        """
        contact_set1 = cls.__group_contacts_by_property_name(from_contacts)
        contact_set2 = cls.__group_contacts_by_property_name(with_contacts)

        contacts = []

        for contact_name in contact_set1:
            # Check whether a contact of this name has been already set as primary
            # in the second list of contacts.  If so, every contact of this name in
            # the first list must be set as not primary.
            is_primary_contact_already_defined = any([_contact.is_primary for _contact in contact_set2[contact_name]])

            # Retrieve the list of contacts from the first list that don't exist in
            # the second list.
            contacts.extend([
                Contact(
                    contact.property_name,
                    contact.property_value,
                    not is_primary_contact_already_defined and contact.is_primary,
                    contact.is_verified)
                for contact in contact_set1[contact_name]
                if not any([
                    True
                    for _contact in contact_set2[contact_name]
                    if contact.property_value == _contact.property_value
                ])
            ])

        return contacts

    def add_contact(
            self,
            account_id,
            contact,
            connection=None,
            context=None,
            has_been_verified=False,
            language=DEFAULT_LOCALE,
            to_be_verified=False,
            verification_code_length=None):
        """
        Add a given contact information to the specified user account as long
        as this contact information has not been verified by another user.

        If this contact information has been verified, the function will
        silently disable this contact information from account of any user who
        would have added it to their account, and cancel any request that
        would have been generated to verify this contact information.


        :param account_id: identification of the account of the user who adds
            his contact information.

        :param contact: An object `Contact`.

        :param connection: An object `RdbmsConnection` with auto commit.

        :param context: A JSON expression corresponding to the context
            in which this contact has been added and need to be verified.
            This parameter is only used if the parameter `to_be_verified` is
            set to `True`.

        :param has_been_verified: Indicate whether this contact information
            has been verified through a challenge/response process.  This
            parameter is intended to be set by the function `confirm_contact`
            only.

        :param language: An object `Locale` representing the language in which
            to write the email verification request in.

        :param to_be_verified: Indicate whether the platform needs to send a
            request to the user to verify his contact information.  This
            argument cannot be set to `True` if the argument `has_been_verified`
            has been set to `True`.

        :param verification_code_length: Number of digits to generate the
            verification code.

        :raise ContactAlreadyUsedException: if one of the contacts has been
            already registered and verified by another user.

        :raise InvalidArgumentException: if this contact information is stated
            as verified while no challenge process has been passed.
        """
        assert not (has_been_verified and to_be_verified), "Conflicted values"

        # Check whether this contact has been already added to this account.
        account = AccountService().get_account(
            account_id,
            check_status=True,
            connection=connection,
            include_contacts=True)

        if contact in account.contacts:
            return

        # Check whether this contact would been added to another user account
        # and whether it would have been verified.
        (is_available, is_verified) = self.is_contact_available(contact, connection=connection)
        if not is_available and is_verified:
            raise self.ContactAlreadyInUseException("Another user has already verified this contact information")

        # Check whether this contact information can been stated verified.
        if contact.is_verified and not has_been_verified and account.account_type not in [
            AccountType.botnet,
            AccountType.sns,
            AccountType.test
        ]:
            raise self.InvalidArgumentException(
                "A contact information cannot be said verified without passing a challenge process")

        # Add this contact information to the specified user account, and
        # generate a verification request if required.
        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            # If verified, disable this contact information from any other user
            # account, and cancel any request that would have been generated to
            # verify this contact information.
            if contact.is_verified:
                connection.execute(
                    """
                    UPDATE
                        account_contact
                      SET
                        object_status = %(OBJECT_STATUS_DISABLED)s
                      WHERE
                        property_name = %(property_name)s
                        AND property_value = %(property_value)s
                        """,
                    {
                        'OBJECT_STATUS_DISABLED': ObjectStatus.disabled,
                        'property_name': contact.property_name,
                        'property_value': contact.property_value
                    })

                connection.execute(
                    """
                    DELETE FROM 
                        account_contact_verification
                      WHERE
                        property_name = %(property_name)s
                        AND property_value = %(property_value)s""",
                    {
                        'property_name': contact.property_name,
                        'property_value': contact.property_value
                    })

            # Set this contact information as the primary if it has been verified
            # and it his user doesn't have any other contact information than this
            # one.
            if contact.is_verified and not contact.is_primary and not account.contacts:
                contact.is_primary = True

            # Add this contact information to the specified user account.
            connection.execute(
                """
                INSERT INTO account_contact(
                    account_id,
                    property_name,
                    property_value,
                    is_primary,
                    is_verified)
                  VALUES (
                    %(account_id)s,
                    %(property_name)s,
                    %(property_value)s,
                    %(is_primary)s,
                    %(is_verified)s
                  )
                """,
                {
                    'account_id': account.account_id,
                    'is_primary': contact.is_primary or False,
                    'property_name': contact.property_name,
                    'property_value': contact.property_value,
                    'is_verified': contact.is_verified or False,
                }
            )

            # If this contact information needs to be verified, generate a
            # verification request that will be sent to the user.
            if to_be_verified and account.account_type == AccountType.standard:
                self.request_contact_verification(
                    contact,
                    account_id=account_id,
                    connection=connection,
                    context=context,
                    language=language,
                    verification_code_length=verification_code_length)

    def confirm_contact(
            self,
            connection: rdbms.RdbmsConnection = None,
            verification_code: str = None,
            app_id: uuid.UUID = None,
            contact: Contact = None,
            request_id: uuid.UUID = None):
        """
        Enable a contact information that a user has confirmed through a
        challenge/response verification process


        There are two methods to complete a contact verification process:

        1. Either to provide the identification of the contact verification
           request `request_id`

        2. Either to provide the activation code that was generated with the
           contact verification request, and to provide the contact that was
           requested to be verified (the activation code might not be unique
           among all the contacts that are requested to be verified).

        If the user doesn't have any account, the function will automatically
        create an account for him and associate this contact information to
        his account.

        The function silently deletes any other verification requests that
        would have been initiated by other users for this same contact
        information.  The function also automatically soft-deletes any user
        account that would be only linked to this contact information.

        If the user doesn't have any primary contact information defined of
        this type, the function will automatically set this contact
        information as the primary.


        :warning: this function MUST NOT be directly surfaced to any client
            application, but SHOULD be used by the server controller of a
            client application.


        :param app_id: identification of the client application such as a Web,
            a desktop, or a mobile application, that accesses the service.

        :param connection: An object `RdbmsConnection` with auto commit.

        :param contact: The contact information that the user is verifying.
            The argument `verification_code` MUST be also passed to this
            function.

        :param request_id: identification of the verification request that has
            been sent to the user.

        :param verification_code: A pseudo-random number (nonce) that was
            generated and sent to the user to verify his contact information.
            The argument `contact` MUST be also passed to this function.


        :return: The function returns an instance `contact` or a tuple of
            instances `contact, session` depending on whether the parameter
            `auto_sign_in` has been set to, respectively `False`, or
            `True`:

             * `contact`: an instance containing the following the members:

                * `account_id` (required): identification of the account of the user
                  who has confirmed this contact information.

                * `locale` (optional): an instance `Locale` representing the
                  preferred language that this user has selected when he has added
                  this contact information to his account, or, if not defined at this
                  time, when he has confirmed this contact information or, if still
                  not defined at this time, the default locale.

                * `name` (required): item of the enumeration `ContactName`.

                * `value`: value of this contact information.

            * `session`: an instance containing the following members;

                * `creation_time` (optional): time when this account has been
                  registered.  This information should be stored by the client
                  application to manage its cache of accounts.

                * `expiration_time` (optional): time when the login session will
                  expire.  This information is provided if the client application
                  requires the platform to automatically sign-in the user (cf.
                  parameter `auto_sigin`).

                * `object_status` (optional): current status of this user account.

                * `session_id` (optional): identification of the login session of
                  the user.  This information is provided if the client application
                  requires the platform to automatically sign-in the user (cf.
                  parameter `auto_sign_in`).


        :raise DeletedObjectException: if the user account has been deleted
            while the argument `check_status` has been set to `True`.

        :raise DisabledObjectException: if the user account has been disabled
            while the argument `check_status` has been set to `True`.

        :raise UndefinedObjectException: if the specified identification
            doesn't refer to a user account registered against the platform.
        """
        if request_id is None and verification_code is None:
            raise self.InvalidArgumentException("Either a request ID or an activation code MUST be passed")

        if verification_code and not contact:
            raise self.InvalidArgumentException("A contact information MUST be passed")

        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            cursor = connection.execute(
                """
                DELETE FROM
                    account_contact_verification
                  WHERE
                    (%(request_id)s IS NOT NULL AND request_id = %(request_id)s)
                    OR (%(verification_code)s IS NOT NULL 
                        AND verification_code = %(verification_code)s
                        AND property_name = %(property_name)s
                        AND property_value = %(property_value)s)
                  RETURNING
                    account_id,
                    language,
                    property_name, 
                    property_value
                """,
                {
                    'verification_code': verification_code,
                    'property_name': contact.property_name,
                    'property_value': contact.property_value,
                    'request_id': request_id,
                })

            row = cursor.fetch_one()

            request = row and row.get_object({
                'account_id': cast.string_to_uuid,
                'language': Locale,
                'property_name': ContactName,
            })

            if request is None:
                raise self.UndefinedObjectException("The contact verification request doesn't exist")

            contact = Contact(request.property_name, request.property_value, is_verified=True)

            # Remove reference to this contact information from any other user
            # account.  Cancel any verification request of this contact
            # information.
            self.__cleanse_contact_references(request.account_id, contact, connection=connection)

            # If this request was linked to a particular user account, enable his
            # contact information.
            if request.account_id:
                self.__enable_account_contact(request.account_id, contact, connection=connection)

            return Object(
                account_id=request.account_id,
                contact=contact,
                language=request.language)

    def find_account(self, contact, connection=None):
        """
        Return the account that matches a contact.


        :param contact: An object `Contact`.

        :param connection: An object `RdbmsConnection`.


        :return: `None` if no account matches the specified contact, or an
            object containing the following attributes:

            * `account_id` (required): Identification of the account.

            * `is_primary` (required): Indicate whether this contact information is
              the primary one for the given property.

            * `is_verified` (required): Indicate whether this contact information
              has been verified.

            * `object_status` (required): Current status of this contact information.
        """
        with self.acquire_rdbms_connection(auto_commit=False, connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT
                    account_id,
                    is_primary,
                    is_verified,
                    object_status
                  FROM
                    account_contact
                  WHERE
                    property_name = %(property_name)s
                    AND property_value = %(property_value)s
                """,
                {
                    'property_name': contact.property_name,
                    'property_value': contact.property_value,
                }
            )

            row = cursor.fetch_one()
            account = row and row.get_object({
                'account_id': cast.string_to_uuid,
                'object_status': ObjectStatus
            })

            return account

    def get_contact_verification_request(self, contact):
        """
        Return the request send to a user to verify his contact information.


        :note: this function MUST not be surfaced to client applications.


        :param contact: an instance `Contact`.


        :return: an instance containing the following members:

            * `account_id` (optional): identification of the account of the user
              who requested to verify his contact information, if he was logged in
              when he provided his contact information.

            * `elapsed_time` (required): elapsed time in second between since
              this contact verification request has been generated.

            * `request_id` (required): identification of the contact verification
              request.
        """
        with self.acquire_rdbms_connection(auto_commit=False) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    request_id,
                    account_id,
                    EXTRACT(EPOCH FROM current_timestamp - update_time) AS elapsed_time
                  FROM
                    account_contact_verification
                  WHERE
                    property_name = %(property_name)s
                    AND property_value = %(property_value)s
                """,
                {
                    'property_name': contact.property_name,
                    'property_value': contact.property_value
                })
            row = cursor.fetch_one()

            if row is None:
                raise self.UndefinedObjectException(
                    "No verification request corresponds to the specified contact")

            return row.get_object({
                'account_id': cast.string_to_uuid,
                'request_id': cast.string_to_uuid
            })

    def get_contacts(
            self,
            account_id,
            connection=None,
            property_name=None):
        """
        Return the list of contact information of an account


        :param account_id: Identification of a user account.

        :param connection: An object `RdbmsConnection`.

        :param property_name: An item of the enumeration `ContactName` to
            filter the list of contact information of the user.


        :return: A list of `Contact` objects.
        """
        with self.acquire_rdbms_connection(connection=connection) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    property_name,
                    property_value,
                    is_primary,
                    is_verified
                  FROM
                    account_contact
                  WHERE
                    account_id = %(account_id)s
                    AND object_status = %(OBJECT_STATUS_ENABLED)s
                    AND (%(property_name)s IS NULL OR property_name = %(property_name)s) 
                """,
                {
                    'OBJECT_STATUS_ENABLED': ObjectStatus.enabled,
                    'account_id': account_id,
                    'property_name': property_name,
                })

            contacts = [
                Contact.from_object(row.get_object())
                for row in cursor.fetch_all()
            ]

            return contacts

    def set_primary_contact(
            self,
            account_id,
            contact,
            add_if_not_exists=False,
            connection=None):
        """
        Set the primary contact information of a users

        This new or the existing contact information MUST have been verified.

        The function sets all the contact information of the same type as
        alternates.


        :param account_id: The identification of the account of a user.

        :param contact: An object `Contact` of the contact information of
            the user to set as primary.  The attribute `is_verified` of this
            object MUST be `True`.

        :param add_if_not_exists: Indicate whether to add this contact
            information if it doesn't not exist.

        :param connection: An object `RdbmsConnection` with auto commit.


        :raise InvalidOperationException: If the contact information has not
            been verified yet.

        :raise UndefinedObjectException: If this contact information has not
            been associated to this user account.
        """
        if not contact.is_verified:
            raise self.InvalidArgumentException("The contact information MUST have been verified")

        # Check whether this contact information is already associated to this
        # user.  If not, possibly add this contact information to the user; if
        # already associated to another user, this would raise an exception.
        contacts = self.get_contacts(
            account_id,
            connection=connection,
            property_name=contact.property_name)

        existing_contact = [_ for _ in contacts if _.property_value == contact.property_value]
        if len(existing_contact) == 0:
            if not add_if_not_exists:
                raise self.UndefinedObjectException('This contact information is not associated to this user account')

            self.add_contact(
                account_id,
                contact,
                connection=connection,
                has_been_verified=contact.is_verified)
        else:
            existing_contact = existing_contact[0]

            # Ignore this request if this contact information is already set as the
            # primary contact for this user.
            if existing_contact.property_value == contact.property_value and existing_contact.is_primary:
                return

            # Check whether the contact information already associated to the user
            # has been verified (requirement for primary contact).
            if not existing_contact.is_verified:
                raise self.InvalidOperationException(
                    "A contact information that was not verified cannot be set as primary")

        # Update all the contact information of this type for this user account,
        # setting this particular contact information as the primary, and the
        # other as alternates (if they are not already alternates).
        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            connection.execute(
                """
                UPDATE 
                    account_contact
                  SET
                    is_primary = (property_value = %(property_value)s),
                    update_time = current_timestamp
                  WHERE
                    account_id = %(account_id)s
                    AND property_name = %(property_name)s
                    AND is_primary <> (property_value = %(property_value)s)
                """,
                {
                    'account_id': account_id,
                    'property_name': contact.property_name,
                    'property_value': contact.property_value,
                })


AccountService = module_util.load_class('majormode.perseus.service.account.account_service.AccountService')



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import contextlib
import json
import logging
import re
import time
import urllib.error
import urllib.parse
import urllib.request
import traceback

from majormode.perseus.agent.email_sender_agent import EmailSenderAgent
from majormode.perseus.constant.contact import ContactName
from majormode.perseus.model.locale import DEFAULT_LOCALE
from majormode.perseus.model.locale import Locale
from majormode.perseus.utils import cast
from majormode.perseus.utils import rdbms


class EmailAddressVerificationAgent(EmailSenderAgent):
    # Regular expression matching the pattern of a placeholder where its
    # name needs to be replaced by its corresponding value.
    #
    # Placeholder names are surrounded by "::" and "::".  Placeholder names
    # must consist of any letter (A-Z), any digit (0-9), an underscore, or
    # a dot.  Placeholder names must start with a letter.
    REGEX_PLACEHOLDER_NAME = re.compile(r'::((?i)[a-z][a-z0-9_.]*)::')

    MAXIMUM_BATCH_SIZE = 1000

    def __init__(
            self,
            smtp_connection_properties,
            rdbms_connection_properties,
            template_url,
            verbose=False):
        super().__init__(smtp_connection_properties)

        self.rdbms_connection_properties = rdbms_connection_properties
        self.template_url = template_url
        self.verbose = verbose

    @classmethod
    def expand_placeholders_value(cls, content, placeholders, ignore_unused_placeholders=False):
        """
        Replace the placeholders defined in the given content with their
        corresponding values.


        :param content: The content with placeholders {{placeholder}} to be
            replaced with their corresponding values.

        :param placeholders: A dictionary where the key corresponds to the
            name of a placeholder, the value corresponding to the value of the
            placeholder.

        :param ignore_unused_placeholders: Indicate whether to ignore
            placeholders that would have been defined but that would have not
            been declared in the given content.


        :return: The content where the placeholders have been replaced with
            their corresponding values.
        """
        # Find the list of placeholder names that have been used in the given
        # document but that have not been defined with a corresponding value.
        used_placeholders = set([
            (match.group(0), match.group(1))
            for match in cls.REGEX_PLACEHOLDER_NAME.finditer(content)
        ])

        used_placeholder_names = [name for (_, name) in used_placeholders]

        if used_placeholders:
            undefined_placeholder_names = [
                name
                for name in used_placeholder_names
                if name not in placeholders
            ]

            assert len(undefined_placeholder_names) == 0, \
                "The following placeholders are declared but have not been defined: " \
                f"{','.join(undefined_placeholder_names)}"

        # Find the list of placeholder names that have been defined but have not
        # been used in the given document.
        if placeholders:
            unused_placeholder_names = [
                name
                for name in placeholders
                if name not in used_placeholder_names
            ]

            if len(unused_placeholder_names) > 0:
                warning_message = "The following placeholders are defined but have not been declared: " \
                    f"{','.join(unused_placeholder_names)}"
                assert ignore_unused_placeholders, warning_message
                logging.warning(warning_message)

        # Replace the placeholders referred in the content by their
        # corresponding value.
        for (placeholder_expression, placeholder_name) in used_placeholders:
            content = content.replace(placeholder_expression, placeholders[placeholder_name])

        return content

    def fetch_html_template(
            self,
            locale=None,
            maximum_attempt_count=3,
            sleep_duration_between_attempts=10):
        """
        Fetch the HTML template from its URL.


        @note: the function doesn't handle cached version of HTML template.
            The function fetches the HTML template URL each the function is
            called.  The reason is the HTML template may be changed and
            caching the HTML template would result in returning a deprecated
            version.


        :param locale: an instance `Locale` that indicates the language of
            the HTML template to fetch should be written in.  If this HTML
            template is available for the specified language, the function
            returns the HTML template written is the closest language to the
            one specified, or the default language.

        :param maximum_attempt_count: maximal number of failed attempts to
            fetch the HTML template URL before the function raises an
            exception.

        :param sleep_duration_between_attempts: indicate the time in seconds
            during which the current thread is suspended when an attempt to
            fetch the HTML template URL fails, before attempting again.


        :return: the content of the HTML template in the specified language,
            or the closest language if the HTML template is not available in
            the specified language.


        :raise Exception: if all the attempts in fetching the HTML template
            URL have failed.
        """
        attempt_count = 0

        while True:
            try:
                opener = urllib.request.build_opener()

                opener.addheaders = [
                    ('User-Agent', 'Mozilla/5.0'),
                    ('Accept-Language', (locale or DEFAULT_LOCALE).to_http_string())
                ]

                with contextlib.closing(opener.open(self.template_url)) as handle:
                    return handle.read().decode()

            except urllib.error.URLError:
                logging.debug(traceback.print_exc())
                logging.error(f"An HTTP error occurred while fetching the URL {self.template_url}")

                attempt_count += 1
                if attempt_count > maximum_attempt_count:
                    raise Exception("All the attempts in fetching the specified URL have failed")

                time.sleep(sleep_duration_between_attempts)

    def fetch_requests(
            self,
            limit=None):
        """
        Return a list of unverified email addresses that their respective
        users are requested to confirm.


        :param limit: maximum number of email address verification requests to
            be returned.


        :return: a list of instance containing the following members:

            * `account_id` (required): identification of the account of the user
              who provided  his email address.

            * `email_address` (required): email_address of the user.

            * `context` (optional): A JSON expression corresponding to the context
              in which this contact has been added and needs to be verified.

            * `creation_time` (required): time when the user provided this email
              address.

            * `locale` (optional): an instance `Locale` referring to the preferred
              language of the user.

            * `request_id` (required): identification of the contact information
              request.
        """
        with rdbms.RdbmsConnection.acquire_connection(
                self.rdbms_connection_properties,
                auto_commit=False) as connection:
            cursor = connection.execute(
                """
                SELECT 
                    request_id,
                    account_id,
                    value AS email_address,
                    locale,
                    context,
                    request_count,
                    attempt_count,
                    creation_time,
                    update_time,
                    attempt_time
                  FROM 
                    account_contact_verification
                  WHERE 
                    (attempt_time IS NULL OR current_timestamp - attempt_time >= '1 day'::interval)
                    AND name = %(name)s
                  ORDER BY 
                    attempt_time ASC NULLS FIRST
                  LIMIT %(limit)s
                """,
                {
                    'limit': min(limit or self.MAXIMUM_BATCH_SIZE, self.MAXIMUM_BATCH_SIZE),
                    'name': ContactName.EMAIL
                })

            requests = [
                row.get_object({
                    'account_id': cast.string_to_uuid,
                    'attempt_time': cast.string_to_timestamp,
                    'context': cast.string_to_json,
                    'creation_time': cast.string_to_timestamp,
                    'locale': Locale,
                    'request_id': cast.string_to_uuid,
                    'update_time': cast.string_to_timestamp
                })
                for row in cursor.fetch_all()
            ]

            return requests

    @staticmethod
    def get_subject(content):
        """
        Return the subject of the HTML email taken from the title of the HTML
        template.


        :return: the subject of the HTML email template.


        :raise AssertError: If the content of the HTML email template has not
            been read, of if this HTML has no title defined.
        """
        start_offset = content.find('<title>')
        assert start_offset > 0, "The HTML template has no title defined"
        start_offset = content.find('>', start_offset) + 1

        end_offset = content.find('</title>', start_offset)
        assert end_offset > 0, "The HTML template has no title end tag."

        return content[start_offset:end_offset]

    def update_request(self, request):
        """
        Update the attributes of the specified email address verification
        request that has been successfully processed.


        :param request: An instance containing the following attributes:

            * `request_id` (required): identification of the request.
        """
        with rdbms.RdbmsConnection.acquire_connection(self.rdbms_connection_properties, auto_commit=True) as connection:
            connection.execute(
                """
                UPDATE account_contact_verification
                  SET attempt_count = attempt_count + 1,
                      update_time = current_timestamp,
                      attempt_time = current_timestamp
                  WHERE request_id = %(request_id)s
                """,
                {
                    'request_id': request.request_id
                })



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime
import functools
import inspect
import json
import logging
import re
import socket
import sys
import time
import traceback
import types
import urllib
import urllib.parse
import uuid
import zlib

from majormode.perseus.constant import regex
from majormode.perseus.constant.http import HttpMethod
from majormode.perseus.constant.http import HttpScheme
from majormode.perseus.constant.http import REGEX_PATTERN_USER_AGENT
from majormode.perseus.constant.stage import EnvironmentStage
from majormode.perseus.model import obj
from majormode.perseus.model.app import ClientApplication
from majormode.perseus.model.obj import Object
from majormode.perseus.model.enum import Enum
from majormode.perseus.model.locale import Locale
from majormode.perseus.model.version import Version
from majormode.perseus.service.application.application_service import ApplicationService
from majormode.perseus.service.account.session_service import SessionService
from majormode.perseus.utils import cast
from majormode.perseus.utils import string_util

import settings


class HttpRequestLog:
    """
    Represent logging information about a HTTP request that a client
    application sent to the RESTful API server application.
    """
    def __init__(
            self,
            request,
            response,
            execution_duration,
            dispatch_duration,
            exception=None):
        """
        Build a `HttpRequestLog` instance providing detailed information
        about the processing of a HTTP request that a client application sent
        to the platform.


        :param request: A `HttpRequest` instance.

        :param response: A `HttpResponse` instance.

        :param execution_duration: The time spent in processing the client's
            HTTP request.

        :param dispatch_duration: The time spent in sending the result to the
            client.

        :param exception: A possible exception that the processing of this
            HTTP request may have raised.
        """
        self.log_id = uuid.uuid1().hex
        self.app_id = str(request.app_id) if hasattr(request, 'app_id') else None
        self.api_version = settings.API_VERSION
        self.host_ip_address = socket.gethostbyname(socket.gethostname())
        self.client_ip_address = request.client_ip_address
        self.http_method = str(request.http_method)
        self.uri = request.uri
        self.arguments = request.arguments
        self.headers = request.headers
        self.request_size = 0 if request.body is None else len(request.body)
        self.response_size = len(str(response))
        self.execution_duration = (execution_duration.seconds * 1000) + (execution_duration.microseconds / 1000)
        self.dispatch_duration = (dispatch_duration.seconds * 1000) + (dispatch_duration.microseconds / 1000)
        self.request_time = time.strftime('%Y-%m-%dT%H:%M:%S.000+00', time.gmtime())

        if settings.LOGGING_LEVEL == logging.DEBUG:
            self.request_content = request.body if len(request.uploaded_files) == 0 else '<<files>>',

        if exception is not None:
            self.exception_class = type(exception).__name__
            self.exception_message = str(exception)
            self.exception_traceback = traceback.format_exc()

        self._tag = self.__class__.__name__


class HttpRequest:
    class HttpRequestException(Exception):
        """
        Base class of exception that the HTTP request handler may raise to
        indicate some unexpected conditions.
        """
        def __init__(self, message=None, payload=None):
            """
            Build a new instance of an exception that indicates an unexpected
            condition while handling a client's request.


            :param message: A humanly-readable message that describes this
                exception instance.

            :param payload: A JSON expression that provides detailed information
                about this exception instance.
            """
            super().__init__(message)
            self.payload = payload

    class DeprecatedApiException(HttpRequestException):
        """
        Signal that the client application has specified a deprecated version
        of the RESTful API while the Python handler of the given HTTP request
        requires a compatible version (cf. parameter `compatibility_required`
        of the decorator `@http_request`).
        """

    class InvalidArgumentException(HttpRequestException):
        """
        Signal that the client application has passed an argument of the wrong
        type or of the wrong value, within the HTTP request to process, as the
        HTTP handler requests (cf. parameter `data_type` of arguments accepted
        by the corresponding HTTP handler).
        """

    class InvalidMessageBodyException(HttpRequestException):
        """
        Indicate that the message body of the HTTP request is not a valid JSON
        expression.
        """

    class MissingArgumentException(HttpRequestException):
        """
        Signal that the application hasn't passed a required argument with the
        HTTP request to process.
        """

    class MissingHeaderException(HttpRequestException):
        """
        Signal that the application hasn't provided a HTTP header required by
        the HTTP request processor.
        """

    class HttpRequestUploadedFile:
        """
        Represent a file that has been uploaded.
        """
        def __init__(self, field_name, file_name, content_type, data):
            """
            Build a new instance of this class.


            :param field_name: Original field name in the form that has been sent
                to the platform to upload a file (cf. multipart MIME data
                stream "multipart/form-data").

            :param file_name: Name of the file as defined in the HTTP request that
                the client application sent to the platform.

            :param content_type: Specify the nature of the data in the content of
                the file that has been uploaded to the platform entity, by
                giving type and subtype identifiers, and by providing auxiliary
                information that may be required for certain types.

            :param data: Content of the file that has been uploaded.
            """
            self._field_name = field_name
            self._file_name = file_name
            self._content_type = content_type
            self._data = data

        @property
        def data(self):
            return self._data

        @property
        def content_type(self):
            return self._content_type

        @property
        def field_name(self):
            return self._field_name

        @property
        def file_name(self):
            return self._file_name

    class ArgumentObject:
        """

        """
        @staticmethod
        def _get_argument(
                payload,
                argument_name,
                data_type=None,
                default_value=None,
                enumeration=None,
                is_required=True,
                item_data_type=None,
                object_class=None):
            """
            Return the value of an argument contained in the JSON expression


            :param payload: A JSON expression.

            :param argument_name: Name of the argument to return the value.

            :param default_value: default value to return if the argument was
                optional and not defined within the HTTP request.

            :param enumeration: An enumeration inheriting from `Enum`, when the
                argument `data_type` is `ArgumentDataType.enumeration`.

            :param is_required: Indicate whether this argument is mandatory, i.e.
                it MUST be provided in the HTTP request, and if not provided, the
                function will raise the exception `MissingArgumentException`.

            :param item_data_type: When the argument to return the value is of
                data type `ArgumentType.list`, the caller can specify the data
                type of the items that this list contains.

            :param object_class: Python class to be used to instantiate a new
                object from the JSON value of the given argument when the
                specified data type of this argument is `object`.  This Python
                class MUST implement a static method `from json` that returns an
                instance of this class providing a JSON expression.


            :return: The casted value of the argument.


            :raise MissingArgumentException: If the JSON expression doesn't
                contain the specified argument while this argument is required.
            """
            if not isinstance(payload, dict):
                raise ValueError("The argument 'payload' MUST be a dictionary")

            if is_required and argument_name not in payload:
                raise HttpRequest.MissingArgumentException(f'Required argument "{argument_name}" is not passed')

            argument_value = payload.get(argument_name) or default_value

            casted_argument_value = HttpRequest.cast_value(
                argument_name,
                argument_value,
                argument_passing=HttpRequest.ArgumentPassing.message_body,
                data_type=data_type or HttpRequest.ArgumentDataType.string,
                enumeration=enumeration,
                item_data_type=item_data_type or HttpRequest.ArgumentDataType.string,
                object_class=object_class)

            return casted_argument_value

        @staticmethod
        def from_json(payload):
            """
            Build an object with a JSON expression


            :param payload: A JSON expression.


            :return: An object built from the JSON expression
            """
            raise NotImplementedError()

    ArgumentDataType = Enum(
        'boolean',
        'date',
        'decimal',
        'dictionary',
        'email_address',
        'enumeration',
        'hexadecimal',
        'integer',
        'ipv4',
        'list',
        'locale',  # RFC 4646
        'macaddr',
        'natural_number',  # Positive integers including zero
        'object',
        'regex',
        'string',
        'time',
        'timestamp',
        'uri',
        'uuid',
        'version'
    )

    ArgumentPassing = Enum(
        'query_string',
        'message_body'
    )

    # Regular expression that matches a valid User-Agent HTTP header.
    REGEX_USER_AGENT = re.compile(REGEX_PATTERN_USER_AGENT)

    def __init__(
            self,
            http_method,
            uri,
            arguments,
            headers,
            remote_ip,
            body=None,
            uploaded_files=None):
        """
        Build an instance `HttpRequest` representing the HTTP request that a
        client application sent to the platform.


        :note: The method `_prepare_` MUST be called to complete the
            instantiation of this object.


        :param http_method: An instance `HttpMethod` that indicates the HTTP
            method of this request.

        :param uri: Uniform Resource Identifier that names a resource of the
            platform the HTTP request requires to gain access to.

        :param arguments: A dictionary of key/value pairs representing the
            arguments passed in the query string of the HTTP request.

        :param headers: A dictionary of key/value pairs representing the
            headers of the HTTP request.

        :param remote_ip: The Internet Protocol address of the client
            application that initiates this HTTP request.

        :param body: Content of the message body of the HTTP request when the
            HTTP method is `POST` or `PUT`.

        :param uploaded_files: A list of instances `HttpRequestUploadedFile`
            mapping the files that the client application has uploaded to
            the platform along the HTTP request.
        """
        if http_method not in HttpMethod:
            raise Exception(f'Unsupported HTTP method {http_method}')

        self.http_method = http_method
        self.uri = uri
        self.arguments = arguments
        self.headers = headers
        self.uploaded_files = uploaded_files or []

        # Retrieve the IP address of the client application that sent this HTTP
        # request.
        x_forwarded_for = headers.get('X-Forwarded-For')
        self.client_ip_address = cast.string_to_ipv4(
            x_forwarded_for.split(',')[0] if x_forwarded_for else headers.get('X-Real-Ip', remote_ip))

        self.request_time = time.strftime('%Y-%m-%dT%H:%M:%S.000+00', time.gmtime())

        # Check whether the client application sends compressed request using
        # "gzip" method.
        #
        # To decompress gzipped data with zlib, the windowBits parameter must
        # be 16 + MAX_WBITS (cf. function `deflateInit2`;
        # http://www.zlib.net/manual.html]::
        #
        #   windowBits can also be greater than 15 for optional gzip encoding.
        #   Add 16 to windowBits to write a simple gzip header and trailer
        #   around the compressed data instead of a zlib wrapper."
        #
        # Note: other technique but somewhat slower:
        #
        #    import StringIO
        #    file_object = StringIO.StringIO(self.request.body)
        #    gzipper = gzip.GzipFile(fileobj=file_object)
        #    data = gzipper.read()gzip
        #
        #
        # @note: We keep the body message even for HTTP method other than POST
        #     and PUT.  There are HTTP requests with method GET that need to
        #     pass sensitive data to check a resource's status (e.g., password
        #     reset request, providing a nonce).  HTTP request's query string
        #     is not encrypted, and therefore passing sensitive data in the
        #     query would allow man-in-the-middle (MITM) attack.
        self.raw_body = None if body is None or len(body) == 0 \
            else zlib.decompress(body, 16 + zlib.MAX_WBITS) if headers.get('Content-Encoding') == 'gzip' \
            else body

        content_type = self.headers.get('Content-Type')
        if content_type:
            content_type_parts = [part.strip() for part in content_type.split(';')]
            media_type, encoding = content_type_parts if len(content_type_parts) == 2 else (content_type_parts[0], None)

        self.body = None if self.raw_body is None \
            else json.loads(self.raw_body) if content_type and media_type == 'application/json' \
            else self.raw_body

    # def _prepare(self):
    #     """
    #     Complete the instantiation of the HTTP request object.  This method
    #     MUST be called once the `HttpRequest` instance has been built.
    #
    #     This function is responsible for uncompressing gzip content of the
    #     HTTP request message body when the HTTP method is `PUT` or `POST`
    #     and the content encoding has been indicated as of `gzip`.
    #
    #     This function is responsible for converting the string content (cf.
    #     member `body`) of the HTTP request's message body into a JSON object
    #     when the content type has been indicated as of `application/json`.
    #
    #     :raise Exception: if decompression issue occurs or if JSON string
    #            representation is not conform to the JSON specification.
    #     """
    #     # Check whether the client application sends compressed request using
    #     # "gzip" method.
    #     #
    #     # To decompress gzipped data with zlib, the windowBits parameter must
    #     # be 16 + MAX_WBITS (cf. function `deflateInit2`;
    #     # http://www.zlib.net/manual.html]::
    #     #
    #     #   windowBits can also be greater than 15 for optional gzip encoding.
    #     #   Add 16 to windowBits to write a simple gzip header and trailer
    #     #   around the compressed data instead of a zlib wrapper."
    #     #
    #     # Note: other technique but somewhat slower:
    #     #
    #     #    import StringIO
    #     #    file_object = StringIO.StringIO(self.request.body)
    #     #    gzipper = gzip.GzipFile(fileobj=file_object)
    #     #    data = gzipper.read()gzip
    #     self.body = None if self.body is None or len(self.body) == 0 \
    #         else zlib.decompress(self.body, 16 + zlib.MAX_WBITS) if self.headers.get('Content-Encoding') == 'gzip' \
    #         else self.body
    #
    #     if self.body is not None and self.headers.get('Content-Type') == 'application/json':
    #         self.body = json.loads(self.body)
    #
    #     return self

    @staticmethod
    def cast_value(
            argument_name,
            argument_value,
            argument_passing,
            data_type=ArgumentDataType.string,
            enumeration=None,
            item_data_type=ArgumentDataType.string,
            object_class=None):
        """
        Convert a string to the specified data type.


        :param argument_name: The name of the argument which value is casted.
            This parameter is only required for debugging purpose.

        :param argument_value: The string to be converted to the specified
            data type.

        :param argument_passing: An item of `HttpRequest.ArgumentPassing`
            that indicates whether the argument value has been passed within
            the query parameters of the request or in the message body.  This
            parameter is only used to parse list of values.

        :param data_type: An item of `HttpRequest.ArgumentDataType` that
            indicates the data type to cast the string to.

        :param enumeration: A Python class inheriting from `Enum` which
            the string MUST be an item of, when the argument `data_type` is
            `ArgumentDataType.string`.

        :param object_class: A Python class to instantiate a new object from
            the JSON string representation, when the argument `data_type` is
            `object`.  This Python class MUST implement a static method
            `from_json` that returns an instance of this class providing a
            JSON expression.

            If this argument is not defined, while the argument `data_type` is
            `object`, the function uses the class `Object` to instantiate a
            new object from the JSON string representation

        :param item_data_type: An item of `ArgumentDataType` that specifies
            the data type to cast each item of the list passed as argument,
            when `data_type` is `ArgumentDataType.list`.


        :return: The value converted from the string with the specified data
            type.


        :raise InvalidArgumentException: If `data_type` is
            `ArgumentDataType.enumeration` but the argument `enumeration` has
            not be passed to this function.

        :raise ValueError: If `argument_value` is not a valid string
            representation of the type specified in `data_type`.
        """
        if data_type == HttpRequest.ArgumentDataType.boolean:
            return cast.string_to_boolean(argument_value, strict=True)

        if data_type == HttpRequest.ArgumentDataType.date:
            return cast.string_to_date(argument_value)

        if data_type == HttpRequest.ArgumentDataType.decimal:
            return float(argument_value)

        if data_type == HttpRequest.ArgumentDataType.email_address:
            if not string_util.is_email_address(argument_value):
                raise ValueError(f"Invalid email address '{argument_value}'")
            return argument_value

        if data_type == HttpRequest.ArgumentDataType.enumeration:
            if enumeration is None:
                raise HttpRequest.InvalidArgumentException(f'Unspecified Python enumeration for parameter "{argument_name}"')
            return cast.string_to_enum(argument_value, enumeration)

        if data_type == HttpRequest.ArgumentDataType.integer:
            return cast.string_to_integer(argument_value, strict=False)

        if data_type == HttpRequest.ArgumentDataType.natural_number:
            return cast.string_to_natural_number(argument_value, strict=False)

        if data_type == HttpRequest.ArgumentDataType.ipv4:
            return cast.string_to_ipv4(argument_value, strict=False)

        if data_type == HttpRequest.ArgumentDataType.list:
            element_values = argument_value if argument_passing == HttpRequest.ArgumentPassing.message_body \
                else json.loads(argument_value) if item_data_type == HttpRequest.ArgumentDataType.object \
                else [
                    urllib.parse.unquote(element.strip())
                    for element in argument_value.split(',')
                    if len(element) > 0]

            # Convert each element of the list to the require data type.
            casted_elements = [
                HttpRequest.cast_value(
                    f'{argument_name}.element',  # Debugging purpose only
                    element_value,
                    argument_passing,  # Same argument passing
                    data_type=item_data_type,
                    enumeration=enumeration,
                    object_class=object_class)
                for element_value in element_values
            ]

            return casted_elements

        if data_type == HttpRequest.ArgumentDataType.locale:
            return cast.string_to_locale(argument_value, strict=False)

        if data_type == HttpRequest.ArgumentDataType.macaddr:
            return cast.string_to_macaddr(argument_value)

        if data_type == HttpRequest.ArgumentDataType.object:
            return (object_class or obj.Object).from_json(argument_value)

        if data_type == HttpRequest.ArgumentDataType.time:
            return cast.string_to_time(argument_value)

        if data_type == HttpRequest.ArgumentDataType.timestamp:
            return cast.string_to_timestamp(argument_value)

        if data_type == HttpRequest.ArgumentDataType.uuid:
            return cast.string_to_uuid(argument_value)

        if data_type == HttpRequest.ArgumentDataType.version:
            return Version(argument_value)

        return argument_value

    def get_argument(
            self,
            name,
            argument_passing=None,
            data_type=ArgumentDataType.string,
            default_value=None,
            enumeration=None,
            is_required=True,
            item_data_type=ArgumentDataType.string,
            object_class=None):
        """
        Return the value of the HTTP request argument specified by its name.


        :param name: name of the argument to return its value.

        :param argument_passing: An item of `ArgumentPassing` that indicates
            how this argument is passed with the HTTP request, within the
            query string or the message body.  The default argument passing
            depends on the HTTP method of the request:

            * `DELETE`: query string

            * `GET`: query string

            * `POST`: message body

            * `PUT`: message body

            When the value of the argument is passed in the message body, the
            function expects the message body to be a JSON dictionary, and the
            argument's name to be a root key of this dictionary.

        :param data_type: data type of the value of this argument as one
            defined in the enumeration `HttpRequest.ArgumentDataType`.  The
            function might performs some check to ensure that the value of this
            argument corresponds to the specified data type.

        :param default_value: default value to return if the argument was
            optional and not defined within the HTTP request.

        :param enumeration: An enumeration inheriting from `Enum`, when the
            argument `data_type` is `ArgumentDataType.enumeration`.

        :param is_required: Indicate whether this argument is mandatory, i.e.
            it MUST be provided in the HTTP request, and if not provided, the
            function will raise the exception `MissingArgumentException`.

        :param item_data_type: When the argument to return the value is of
            data type `ArgumentType.list`, the caller can specify the data
            type of the items that this list contains.

        :param object_class: Python class to be used to instantiate a new
            object from the JSON value of the given argument when the
            specified data type of this argument is `object`.  This Python
            class MUST implement a static method `from json` that returns an
            instance of this class providing a JSON expression.


        :return: the value of the specified argument as passed with the HTTP
            request.


        :raise InvalidMessageBodyException: if the HTTP method is `POST` or
            `PUT`, but the message body of this request is either empty or it
            is not a valid JSON expression.

        :raise MissingArgumentException: if no argument corresponds to the
            specified name while this argument is required.
        """
        # Determine the argument passing when not specified, depending on the
        # HTTP method.
        if argument_passing is None:
            # Check whether the client application requests the RESTful API server
            # to override the method specified in the request by whether passing the
            # HTTP header `X-HTTP-Method-Override`, whether passing the argument
            # `_method` (a convention becoming increasingly common in other
            # frameworks).
            overriding_http_method = self.get_header(
                'X-HTTP-Method-Override',
                data_type=HttpRequest.ArgumentDataType.enumeration,
                enumeration=HttpMethod,
                is_required=False) \
                    or cast.string_to_enum(
                        self.arguments.get('_method'),
                        HttpMethod,
                        strict=False)

            argument_passing = HttpRequest.ArgumentPassing.message_body \
                if overriding_http_method or self.http_method in (HttpMethod.POST, HttpMethod.PUT) \
                else HttpRequest.ArgumentPassing.query_string

        # Retrieve the value of the argument from either the query string or
        # the message body, depending on the argument passing.
        is_argument_defined = False

        if argument_passing == HttpRequest.ArgumentPassing.query_string:
            if name in self.arguments:
                is_argument_defined = True
                value = self.arguments.get(name)

                # Strings may be UTF-8 encoded bytes escaped with URL quoting.
                if value and data_type == self.ArgumentDataType.string:
                    value = urllib.parse.unquote(value)

        elif self.body:
            if not isinstance(self.body, dict):
                raise HttpRequest.InvalidMessageBodyException()

            is_argument_defined = name in self.body
            if is_argument_defined:
                value = self.body.get(name)

        # Check whether this argument is required while the client application
        # didn't pass this argument with the HTTP request.
        if is_required and not is_argument_defined:
            get_logger().debug(f'Required argument "{name} is not passed')
            raise HttpRequest.MissingArgumentException(f'Required argument "{name}" is not passed')

        # Use the default value is the client application didn't pass the
        # argument in the HTTP request, of if the value of this argument is
        # null.
        return default_value if not is_argument_defined or value is None else \
            HttpRequest.cast_value(
                name,
                value,
                argument_passing,
                data_type=data_type,
                item_data_type=item_data_type,
                object_class=object_class,
                enumeration=enumeration)

    def get_header(
            self,
            field_name,
            data_type=ArgumentDataType.string,
            default_value=None,
            enumeration=None,
            is_required=True,
            item_data_type=ArgumentDataType.string):
        """
        Return the field value of the HTTP header specified by its field name.

        :param field_name: name of the HTTP header's field.

        :param data_type: data type of the value of this argument as one
           defined in the enumeration `HttpRequest.ArgumentDataType`.  The
           function performs some checks to ensure that the value of this
           argument corresponds to the specified data type.

        :param default_value: default value to return if the header was
           optional and not defined within the HTTP request.

        :param enumeration: An enumeration inheriting from `Enum`, when the
            argument `data_type` is `ArgumentDataType.enumeration`.

        :param is_required: Indicate whether this header is mandatory, i.e.
            it MUST be provided in the HTTP request, and if not provided, the
            function will raise the exception `MissingHeaderException`.

        :param item_data_type: When the header to return the value is of
            data type `ArgumentType.list`, the caller can specify the data
            type of the items that this list contains.


        :return: the value of the HTTP header's field.


        :raise MissingHeaderException: the specified field name doesn't
               correspond to any HTTP header defined in this request.
        """
        if field_name not in self.headers:
            if is_required:
                get_logger().debug(f'Required HTTP header "{field_name}"" is not passed')
                raise HttpRequest.MissingHeaderException(f'Required HTTP header "{field_name}" is not passed')
            else:
                return default_value

        return HttpRequest.cast_value(
            field_name,
            self.headers[field_name],
            None,
            data_type=data_type,
            enumeration=enumeration,
            item_data_type=item_data_type)


class HttpMethodSpec:
    """
    Specification of a HTTP method that handles a Uniform Resource
    Identifier (URI) pattern for a given HTTP method (alias an endpoint).
    """
    class HttpMethodSpecException(Exception):
        """
        Signal a bad declaration of the handler of a Uniform Resource
        Identifier endpoint.
        """

    class AlreadyDeclaredURLBitException(HttpMethodSpecException):
        """
        Signal that a URL bit has been declared more than one time in the the
        Uniform Resource Identifier (URI) pattern of a HTTP method handler.
        """
        def __init__(
                self,
                parameter_name,
                uri_pattern,
                handler_module_name,
                handler_class_name,
                handler_method):
            super().__init__(
                f'URL bit "{parameter_name}" declared twice in the endpoint URL "{uri_pattern}" '
                f'({handler_module_name}.{handler_class_name}.{handler_method.__name__})')

    class UnsupportedURLBitDataTypeException(HttpMethodSpecException):
        """
        Signal that an URL bit, declared in the Uniform Resource Identifier
        (URI) pattern of a RESTful API endpoint, is defined with an
        unsupported data type.
        """
        def __init__(
                self,
                parameter_name,
                parameter_type,
                uri_pattern,
                handler_module_name,
                handler_class_name,
                handler_method):
            super().__init__(
                f'Unknown or unsupported data type "{parameter_type}" of the URL bit "{parameter_name}" '
                f'declared in the endpoint URL "{uri_pattern}" '
                f'({handler_module_name}.{handler_class_name}.{handler_method.__name__})')

    # Regular expression that matches an in-line parameter (alias URL bit)
    # defined in a Uniform Resource Identifier (URI) of a RESTful API
    # endpoint.
    #
    # An URL bit is defined of a name and a possible data type, separated
    # by a colon character ":", the whole expression between parentheses.
    #
    # For examples:
    #
    # * (username)
    # * (age:int)
    # * (account_id:uuid)
    REGEX_URL_BIT_PARAMETER = re.compile(r'\(([a-z_]+)(:([0-9a-z_]+)){0,1}\)')

    # Regular expression patterns that matches different parameter types.
    PARAMETER_TYPE_REGEX_PATTERNS = {
        HttpRequest.ArgumentDataType.integer: regex.REGEX_PATTERN_INTEGER,
        HttpRequest.ArgumentDataType.ipv4: regex.REGEX_PATTERN_IPV4,
        HttpRequest.ArgumentDataType.list: r'(([^/,]+,)*|[^/,]+)+',
        HttpRequest.ArgumentDataType.natural_number: regex.REGEX_PATTERN_NATURAL_NUMBER,
        HttpRequest.ArgumentDataType.string: r'[^/]+',
        HttpRequest.ArgumentDataType.uuid: regex.REGEX_PATTERN_UUID,
    }

    def __init__(
            self,
            http_method,
            uri_pattern,
            handler_module_name,
            handler_class_name,
            handler_method):
        """
        Build a new instance of `HttpMethodSpec`.


        :param http_method: One of the supported HTTP 1.1 common methods as
           defined in the enumeration `HttpMethod` of the class `HttpRequest`:

           * `HttpMethod.DELETE`

           * `HttpMethod.GET`

           * `HttpMethod.POST`

           * `HttpMethod.PUT`


        :param uri_pattern: Uniform Resource Identifier (URI) pattern (i.e.,
            a regular expression) that the wrapped method is responsible to
            handle.

        :param handler_module_name: Name of the Python module that contains
            the code of the HTTP request handler.

        :param handler_class_name: Name of the Python class that implements
            the HTTP request handler.

        :param handler_method: The Python function that handles the URI
            pattern.  This function MUST be a method of a class inheriting
            from `HttpRequestHandler`; because, at the time a function is
            defined, it is just a plain function, i.e., it is not bound to any
            class, therefore this requirement cannot be programmatically
            checked.


        :raise AlreadyDeclaredURLBitException: if an URL bit is declared more
            than one time in the specified Uniform Resource Identifier (URI)
            pattern.

        :raise AssertError: If the argument `http_method` is not an item of
            the enumeration `HttpMethod`.

        :raise UnsupportedURLBitDataTypeException: If an in-line parameter
            (i.e., URL bits), declared in the Uniform Resource Identifier (URI)
            pattern, has a data type that is not supported.
        """
        assert http_method in HttpMethod, 'Invalid HTTP method specified'

        self.http_method = http_method

        # Rewrite the specified URI of the RESTful API endpoint to a regular
        # expression where the data type of each declared URL bit has been
        # converted to a matching regular expression.
        self.uri_pattern, self.url_bits = self._rewrite_uri_pattern_with_regex(
            uri_pattern,
            handler_module_name,
            handler_class_name,
            handler_method)

        self.uri_pattern_regex = re.compile(self.uri_pattern)

        self.handler_module_name = handler_module_name
        self.handler_class_name = handler_class_name
        self.handler_method = handler_method

        self.handler_module = None  # Lazy evaluation
        self.handler_class = None  # Lazy evaluation

    @classmethod
    def _rewrite_uri_pattern_with_regex(
            cls,
            uri_pattern,
            handler_module_name,
            handler_class_name,
            handler_method):
        """
        Rewrite the Uniform Resource Identifier (URI) pattern of a RESTful API
        endpoint, where the declared type of each URL bit has been converted
        to the matching regular expression.


        :param uri_pattern: Uniform Resource Identifier (URI) pattern (i.e.,
            a regular expression) that the wrapped method is responsible to
            handle.

        :param handler_module_name: Name of the Python module that contains
            the code of the HTTP request handler.

        :param handler_class_name: Name of the Python class that implements
            the HTTP request handler.

        :param handler_method: The Python function that handles the URI
            pattern.  This function MUST be a method of a class inheriting
            from `HttpRequestHandler`; because, at the time a function is
            defined, it is just a plain function, i.e., it is not bound to any
            class, therefore this requirement cannot be programmatically
            checked.


        :return: the URI pattern with URL bits declared as named groups with
            their respective data type converted to regular expressions, and a
            dictionary of the URL bits and their respective data type (item of
            `HttpRequest.ArgumentDataType`).
        """
        # Retrieve each in-line parameters (i.e., URL bits), and store their
        # type into a dictionary.  Then rewrite the specified URI pattern with
        # named group and the regular expression corresponding to each parameter
        # type.
        url_bits = dict()

        uri_pattern_with_url_bit_named_groups = []
        end_index = 0

        for match in cls.REGEX_URL_BIT_PARAMETER.finditer(uri_pattern):
            parameter_name, _, parameter_type_str = match.groups()

            # Convert the string representation of the parameter data type to an
            # item of the enumeration `ArgumentDataType`.
            try:
                parameter_type = cast.string_to_enum(
                        parameter_type_str,
                        HttpRequest.ArgumentDataType,
                        default_value=HttpRequest.ArgumentDataType.string,
                        strict=False)

            except ValueError:  # Not an item of `HttpRequest.ArgumentDataType`
                exception = cls.UnsupportedURLBitDataTypeException(
                    parameter_name,
                    parameter_type_str,
                    uri_pattern,
                    handler_module_name,
                    handler_class_name,
                    handler_method)

                get_logger().debug(str(exception))
                raise exception

            # Check whether the specified parameter type is supported for URL bit
            # parameter.
            if parameter_type not in cls.PARAMETER_TYPE_REGEX_PATTERNS:
                exception = cls.UnsupportedURLBitDataTypeException(
                    parameter_name,
                    parameter_type_str,
                    uri_pattern,
                    handler_module_name,
                    handler_class_name,
                    handler_method)

                get_logger().debug(str(exception))
                raise exception

            # Check whether this URL bit parameter has been already declared, and
            # if not, store its declaration.
            if parameter_name in url_bits:
                raise cls.AlreadyDeclaredURLBitException(
                    parameter_name,
                    uri_pattern,
                    handler_module_name,
                    handler_class_name,
                    handler_method)

            url_bits[parameter_name] = parameter_type

            # Add the part of the Uniform Resource Identifier (URI) pattern between
            # the previous and the current URL bit declarations.
            start_index = match.start(0)
            uri_pattern_with_url_bit_named_groups.append(uri_pattern[end_index:start_index])

            # Add the current URL bit in the form of a regular expression named
            # group.
            parameter_type_regex = cls.PARAMETER_TYPE_REGEX_PATTERNS[parameter_type]
            uri_pattern_with_url_bit_named_groups.append(f'(?P<{parameter_name}>{parameter_type_regex})')

            end_index = match.end(0)

        # Add the last part of the Uniform Resource Identifier (URI) pattern
        # remaining after the last URL bit declaration.
        uri_pattern_with_url_bit_named_groups.append(uri_pattern[end_index:])

        return ''.join(uri_pattern_with_url_bit_named_groups), url_bits

    def __eq__(self, other):
        """
        Indicate whether this instance is equivalent to another.


        :param other: Another instance of the class `HttpMethodSpec`.


        :return: `True` if the two instances are equivalent; `False` otherwise.
        """
        return isinstance(other, HttpMethodSpec) and \
            self.http_method == other.http_method and \
            self.uri_pattern == other.uri_pattern

    def __hash__(self):
        """
        Return the hash value of the HTTP endpoint method.


        :return: An integer corresponding to the hash value of the HTTP
            endpoint method.
        """
        return hash((self.http_method, self.uri_pattern))

    def process(self, request):
        """
        Call the handler of a Uniform Resource Identifier (URI) endpoint
        providing a HTTP request sent by a client application


        :param request: An instance of `HttpRequest`.


        :return: The result returned by the handler that processed this
            request.
        """
        # Load the Python module of the handler if not already done.
        if self.handler_module is None:
            self.handler_module = sys.modules[self.handler_module_name]
            self.handler_class = getattr(self.handler_module, self.handler_class_name)

        # Build a new instance of the handler's class and access the Python
        # method of this handler.
        handler_class_instance = self.handler_class()
        handler_method = getattr(handler_class_instance, self.handler_method.__name__)

        # Retrieve the path and the query string of the URI endpoint.
        request_path = urllib.parse.urlsplit(request.uri)[2]

        # Build the Python keyworded arguments to pass to the handler's method,
        # if any URL bits declared in the URI endpoint.
        kwargs = {} if len(self.url_bits) == 0 else \
            dict([
                (
                    name,
                    HttpRequest.cast_value(name, value, None, data_type=self.url_bits[name])
                )
                for name, value in self.uri_pattern_regex.match(request_path).groupdict().items()
            ])

        # Execute the handler method passing:
        #
        # * the instance of the handler class as `self`.
        # * the instance `request` as the first always required argument
        # * the keyworded arguments corresponding to URL bit values.
        request._execution_start_time = datetime.datetime.now()
        response = HttpResponse(handler_method(request, **kwargs))
        request._execution_end_time = datetime.datetime.now()

        return response


class HttpRequestHandler:
    """
    Base class of the HTTP request handler that each RESTful API class
    MUST inherit from.
    """
    # Class member that contains all the `HttpMethodSpec` instances that
    # the RESTful API application has successfully loaded.
    #
    # The key corresponds to the hash of `HttpMethodSpec` instance while
    # the value corresponds to the instance itself.  The key is mostly used
    # by the code responsible for adding declared RESTful methods from
    # Python modules, and making sure that two or more methods are declared
    # for the same endpoint (HTTP method and URI(.
    handlers = dict()


class HttpResponse:
    """
    Represent the response to a HTTP request that has been processed by
    the method of RESTful API service.

    It embeds an object that is serialized to a JSON form to be returned
    to the client application that sent the HTTP request.
    """
    class HttpResponseException:
        """
        Base class of exception that the processor of HTTP request raises when
        an exceptional condition occurred while processing this HTTP request.
        """
        def __init__(self, exception):
            """
            Build a new instance `HttpResponseException` providing the inner
            exception raised by the handler method of a RESTful API service that
            processed the HTTP request.


            :param exception: The inner exception raised by the handler method of
                a RESTful API service that processed the HTTP request.
            """
            self.error = exception.__class__.__name__
            self.message = str(exception)
            self.payload = getattr(exception, 'payload', None)

    def __init__(self, result=None):
        """
        Build a new instance `HttpResponse` corresponding to the result of a
        HTTP request returned by the RESTful API service method that has
        processed this request.


        :param result: result returned by the RESTful API service method that
            has processed a HTTP request.
        """
        # Convert the result to an empty dictionary if the result is null, or
        # a dictionary with a unique entry `data` if the result is not a
        # Python built-in collections, nor an object having `__dict__` attribute.
        self._result = dict() if result is None \
            else result if (isinstance(result, (list, set, tuple)) or hasattr(result, '__dict__')) \
            else dict(data=result)

    def __str__(self):
        """
        Return a string representation of the JSON expression of the result
        returned by the RESTful API service method that has processed a HTTP
        request.


        :return: A string representation of the JSON expression of the result.
        """
        return json.dumps(obj.stringify(self._result, trimmable=True))

    @staticmethod
    def from_exception(exception):
        """
        Build a HTTP response from an exception that occurred while processing
        a HTTP request.


        :param exception: An exception raised by the RESTful API service
            method that has processed a HTTP request that failed.


        :return: An instance `HttpResponse` that embeds an instance
            `HttpResponseException` built with the inner exception raised by
            the RESTful API service method that has processed a HTTP request.
        """
        return HttpResponse(HttpResponse.HttpResponseException(exception))

    @property
    def result(self):
        return self._result


def get_logger():
    """
    Return the logger specified in the settings of the RESTful API
    application's settings.


    :return: an instance `Logger`.
    """
    return logging.getLogger(getattr(settings, 'LOGGER_NAME'))


def http_request(
        uri_pattern,
        http_method=HttpMethod.GET,
        authentication_required=False,
        compatibility_required=False,
        json_data=True,
        sensitive_data=False,
        signature_required=True):
    """
    Decorator used to declare a method of HTTP request handler responsible
    for processing a Uniform Resource Identifier (URI) endpoint defined
    with a specific HTTP method (e.g., `GET`).

    This decoration extends the behavior of the wrapped method, both at
    compile time and run time:

    * At compile time, this decorator registers the wrapped method as the
      handler for the specified Uniform Resource Identifier (URI) pattern
      corresponding to a regular expression and a HTTP method such as
      `GET`, `POST`, `PUT`, and `DELETE`.  It checks that no other
      URI pattern has been register for the specified HTTP method.

    * At run time, this decorator performs around processing of the HTTP
    request for which the wrapped method is invoked.  For instance:

      * Ensure that the HTTP request has been correctly signed (cf.
        argument `signature_enabled`)

      * Ensure that the HTTP request transports the identification of the
        session of a user on behalf of the wrapped method is called (cf.
        argument `authentication_required`)

      * Check that the client application is compatible with the version
        of the server platform's API this client application is
        communicating with.


    :param uri_pattern: Uniform Resource Identifier (URI) pattern,
        corresponding to a regular expression, the wrapped method is
        responsible to handle.

    :param http_method: An item of the enumeration `HttpMethod`.  The
        selected HTTP method indicates the desired action to be performed
        on the resource identified by the URI.  What this resource
        represents depends on the implementation of the method wrapped.

    :param authentication_required: Indicate whether the wrapped method
        requires the user, on behalf of whom the HTTP request is sent, to
        be authenticated or not.  If `True`, the client application MUST
        provide the HTTP header field `X-Authentication`, passing the
        identification of the login session of this user.

    :param compatibility_required: Indicate whether the wrapped method
        requires the client application to be strictly compatible with
        the version of the RESTful API application's.  If `True`, the
        client application MUST provide the HTTP header field
        `X-API-Version` which value MUST comply with a version number
        using a standard tuple of integers: `major.minor.patch`.

    :param json_data: Indicate whether the message body of HTTP methods
        such as `POST` and `PUT` MUST be a string representation of a
        JSON expression.

    :param sensitive_data: Indicate whether sensitive data could be
        transmitted back and forth among the HTTP request or its response
        between the client application and the RESTful API application.

        Sensitive data correspond to a class of data, including personal,
        financial, legal information, which should be carefully handled,
        such as, for instance:

        * user account password
        * Personal Identification Number (PIN)
        * Social Security number (SSN)
        * credit card number or banking information
        * passport number
        * tax information
        * credit reports

        If `True`, the RESTful API application framework doesn't trace
        any data that are transmitted by the client application or
        returned by the RESTful API application, for environment stages
        other than development, integration, and test.

    :param signature_required: indicate whether the wrapped method
        requires the HTTP request to be signed with the Secret Key tied
        with the Consumer Key of the client application that sent this
        request.  If `True`, the client application MUST provide the
        HTTP header fields `X-API-Key` and `X-API-Sig`.


    :return: A function that decorates the given method.


    :raise Exception: If some unexpected conditions occurred while
        decorating the HTTP request handler, such as the URI pattern has
        been already registered with another method, or some options
        passed to this decorator have incoherent values, etc.
    """
    # Check that the method decorated is bounded to a class that inherits
    # from `HttpRequestHandler`.
    #
    # @note: it does seem possible to get the class of the method that is
    # decorated such as:
    #
    #     class_name = inspect.getouterframes(inspect.currentframe())[1][3]
    #     current_module = sys.modules[globals()['__name__']]
    #     _class_ = getattr(current_module, class_name)
    #     if not issubclass(_class_, HttpRequestHandler):
    #        raise Exception('The decorator %s MUST be used for a method of an instance of %s' %\
    #            (inspect.getframeinfo(inspect.currentframe()).function, HttpRequestHandler.__name__))
    #
    # When the decorator is evaluated, the class of the method is not
    # yet defined.
    #
    #     class MyHttpRequestHandler(HttpRequestHandler):
    #         @http_request_handler(uri='^foo$')
    #         def foo(self):
    #             pass
    #
    # meaning that the code:
    #
    #      _class_ = getattr(current_module, class_name)
    #
    # will raise an error as the specified class is not yet bounded to the
    # module.

    # Determine the name of the class of the method that is decorated, which
    # is a bit tricky as a function only becomes a method at runtime.  At
    # the time a function is defined, it is just a plain function, it is not
    # bound to any class.  However, when the decorator is called, even
    # though the function is not yet bound to the class, it is possible to
    # inspect the stack and discover the string name of the class that is
    # calling the decorator.
    class_name = inspect.getouterframes(inspect.currentframe())[1][3]

    def wrapper(method):
        http_method_spec = HttpMethodSpec(http_method, uri_pattern, method.__module__, class_name, method)

        # Check that the specified HTTP method for the given Uniform Resource
        # Identifier (URI) has not been defined for another HTTP request handler.
        existing_http_method_spec = HttpRequestHandler.handlers.get(hash(http_method_spec))
        if existing_http_method_spec:
            logger = get_logger()
            logger.error(f'Multiple handler declarations for endpoint "{http_method} {uri_pattern}":')
            logger.error('- Existing endpoint: {}.{}.{}'.format(
                existing_http_method_spec.handler_module_name,
                existing_http_method_spec.handler_class_name,
                existing_http_method_spec.handler_method.__name__))
            logger.error('- New endpoint: {}.{}.{}'.format(method.__module__, class_name, method.__name__))
            raise Exception(f'The endpoint {http_method} {uri_pattern} is already registered')

        # Check coherence of the options passed to the decorator.
        if authentication_required and not signature_required:
            logger = get_logger()
            logger.error('Method {}.{}.{} requires authentication without signature'.format(
                method.__module__, class_name, method.__name__))
            raise Exception('Signature MUST be requested when authentication is required')

        # Register the specified class method to handle the specified HTTP
        # method for the given Uniform Resource Identifier (URI).
        HttpRequestHandler.handlers[hash(http_method_spec)] = http_method_spec

        @functools.wraps(method)
        def decorated(self, request, *args, **kwargs):
            # Retrieve the identification of the RESTful API application that the
            # HTTP request is sent to.  Check the signature of the HTTP request when
            # required.
            consumer_key = request.get_header('X-API-Key', data_type=HttpRequest.ArgumentDataType.hexadecimal)
            if signature_required or consumer_key:
                request.app_id = get_request_application_id(request, signature_required)

            # Retrieve the version of the RESTful API that the client application
            # has integrated, and detailed information about this client
            # application if identified.
            request.client_api_version, request.client_application = get_client_application_info(
                request,
                compatibility_required
            )

            # Retrieve the login session of the user on behalf of whom the HTTP
            # request is sent to the RESTful API server.
            session = get_account_session(request, request.app_id, authentication_required)
#            request.account_id = session and session.account_id
            request.session = session

            # The content type of HTTP POST and HTTP PUT requests MUST be of
            # "application/json", unless the HTTP request corresponds to file(s)
            # upload, in which case the content type of the HTTP request MUST be of
            # "multipart/form-data; (...)".
            if json_data and request.http_method in (HttpMethod.POST, HttpMethod.PUT):
                content_type = request.headers.get('Content-Type')
                expected_content_type = 'multipart/form-data' if len(request.uploaded_files) > 0 else 'application/json'
                if content_type is None or content_type.split(';')[0] != expected_content_type:
                    raise Exception(f'The content type of this HTTP request is expected to be "{expected_content_type}"')

            # Execute the method responsible for handling the HTTP request.
            return method(self, request, *args, **kwargs)

        return decorated

    return wrapper


def get_account_session(request, app_id, authentication_required):
    """
    Return the login session of the user on behalf whom a HTTP request is
    sent to the RESTful API server.


    :param request: An instance `HttpRequest`.

    :param app_id: Identification of the client application.

    :param authentication_required: Indicate whether the wrapped method
        requires the user, on behalf of whom the HTTP request is sent, to
        be authenticated or not.  If `True`, the client application MUST
        provide the HTTP header field `X-Authentication`, passing the
        identification of the login session of this user.


    :return: An instance `Session` if the user is authenticated against
        the RESTful API server, `None` otherwise.


    :raise ExpiredSessionException: if the account session of the user has
        expired.  The user needs to login again.

    :raise MissingHeaderException: If the request requires the user, on
        behalf of whom the HTTP request is sent, to be authenticated while
        the client application didn't pass the HTTP header
        `X-Authentication`.
    """
    session_id = request.get_header(
        'X-Authentication',
        data_type=HttpRequest.ArgumentDataType.uuid,
        is_required=authentication_required)

    # Retrieve the user session of the user if it exists and is not expired.
    return session_id and SessionService().get_session(app_id, session_id)


def get_client_application_info(request, compatibility_required):
    """
    Return the version of the RESTful API integrated in the client
    application that sent a HTTP request to the RESTful API server.


    :param request: An instance `HttpRequest`.

    :param compatibility_required: Indicate whether the wrapped method
        requires the client application to be strictly compatible with
        the version of the RESTful API application's.  If `True`, the
        client application MUST provide the HTTP header field
        `X-API-Version` which value MUST comply with a version number
        using a standard tuple of integers: `major.minor.patch`.


    :return: A named tuple `(client_api_version, client_application)`
        where:

        * `client_api_version`: An instance `Version` that represents the
          version of the RESTful API integrated in the client application.

        * `client_application`: An instance `ClientApplication` that
          represents information about the client application that sent the
          HTTP request, or `None` if the application is not identified.


    :raise HttpRequest.DeprecatedApiException: If API version
        compatibility is required while the client application has
        integrated a version of the RESTful API older than the one
        currently deployed.
    """
    # When the API version compatibility is required, check the presence of
    # the HTTP header `X-API-Version` and check whether its value is above
    # the version of the RESTful API currently deployed.
    client_api_version = request.get_header(
        'X-API-Version',
        data_type=HttpRequest.ArgumentDataType.version,
        is_required=compatibility_required)

    if compatibility_required:
        if settings.API_VERSION > client_api_version:
            raise HttpRequest.DeprecatedApiException(
                f'Client application deprecated API version {request.api_version} ({settings.API_VERSION} required)')

    # If a User-Agent HTTP header is provided, decompose each component.
    user_agent = request.get_header('User-Agent', is_required=False)
    match = user_agent and HttpRequest.REGEX_USER_AGENT.match(user_agent)

    return client_api_version, match and ClientApplication(
        match.group(1),  # Product name
        Version(match.group(2)),  # Product version
        match.group(3),  # OS name
        Version(match.group(4)),  # OS version
        match.group(5))  # Device model


def get_request_application_id(request, signature_required):
    """
    Return the identification of the client application that sent the
    HTTP request.


    :param request: An instance `HttpRequest`.

    :param signature_required: indicate whether the wrapped method
        requires the HTTP request to be signed with the Secret Key tied
        with the Consumer Key of the client application that sent this
        request.  If `True`, the client application MUST provide the
        HTTP header fields `X-API-Key` and `X-API-Sig`.


    :return: identification of the RESTful API application the HTTP
        request is sent to.
    """
    # When a RESTful API application is deployed on a development or an
    # integration, the signature of an HTTP request is not absolutely
    # required.
    if signature_required:
        environment_stage = getattr(settings, 'ENVIRONMENT_STAGE', EnvironmentStage.dev)
        signature_required = environment_stage not in (
            EnvironmentStage.dev,
            EnvironmentStage.int
        )

    # When the signature of the HTTP request is required, check the presence
    # of the two HTTP headers `X-API-Key` and `X-API-Sig`, and check whether
    # the digest hash of the HTTP request as provided by the client
    # application is valid.
    #
    # @note: Authenticated HTTP request required HTTP signature, therefore
    #     checking the signature of a HTTP request has to be done before the
    #     authentication verification.
    consumer_key = request.get_header('X-API-Key', data_type=HttpRequest.ArgumentDataType.hexadecimal)
    client_signature = request.get_header('X-API-Sig', is_required=signature_required)

    if signature_required or client_signature:
        return ApplicationService().validate_signature(
            consumer_key,
            request.uri,
            request.raw_body if request.headers.get('Content-Type').startswith('application/json') else None,
            client_signature,
            strict=signature_required)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2016 Majormode.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Majormode or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with
# the terms of the license agreement or other applicable agreement you
# entered into with Majormode.
#
# MAJORMODE MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
# SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
# BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  MAJORMODE
# SHALL NOT BE LIABLE FOR ANY LOSSES OR DAMAGES SUFFERED BY LICENSEE
# AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
# DERIVATIVES.
#
# @version $Revision$


# GeoLite2 databases are free IP geolocation databases comparable to,
# but less accurate than, MaxMind’s GeoIP2 databases. GeoLite2 databases
# are updated on the first Tuesday of each month.
#
# IP geolocation is inherently imprecise. Locations are often near the
# center of the population.  Any location provided by a GeoIP database
# should not be used to identify a particular address or household.
#
# Use the Accuracy Radius as an indication of geolocation accuracy for
# the latitude and longitude coordinates we return for an IP address.
# The actual location of the IP address is likely within the area
# defined by this radius and the latitude and longitude coordinates.
#
# The GeoLite2 databases are distributed under the Creative Commons
# Attribution-ShareAlike 4.0 International License.  The attribution
# requirement may be met by including the following in all advertising
# and documentation mentioning features of or use of this database::
#
#     This product includes GeoLite2 data created by MaxMind, available
#     from <a href="http://www.maxmind.com">http://www.maxmind.com</a>.

from majormode.utils import file_util
from majormode.utils import zip_util

import argparse
import contextlib
import os
import re
import urllib2
import zipfile


# URL of the ZIP archive of GeoLite2 City database, which allows to
# determine the country, subdivisions, city, and postal code associated
# with IPv4 and IPv6 addresses worldwide.
GEOLITE2_CITY_ZIP_URL = 'http://geolite.maxmind.com/download/geoip/database/GeoLite2-City-CSV.zip'

# # URL of the ZIP archive of GeoLite Country database, which allows to
# # determine an Internet visitor's country based on their IP address.
# GEOLITE2_COUNTRY_ZIP_URL = 'http://geolite.maxmind.com/download/geoip/database/GeoLite2-Country-CSV.zip'


# REGEX_GEOIP_ZIP_FILE_NAME = re.compile(r'GeoLite2-City-CSV_(\d{8}).zip')



class GeoIP2BlockFile(object):
    # Name of the GeoIP2 block CSV file.
    GEOIP_BLOCK_CSV_FILE_NAME = 'GeoLite2-City-Blocks-IPv4.csv'

    # This is the IPv4 or IPv6 network in CIDR format such as “2.21.92.0/29”
    # or “2001:4b0::/80”.
    FIELD_NAME_NETWORK = 'network'

    # A unique identifier for the network’s location as specified by
    # GeoNames (http://www.geonames.org/).  This ID can be used to look up
    # the location information in the Location file.
    FIELD_NAME_GEONAME_ID = 'geoname_id'

    # The registered country is the country in which the ISP has registered
    # the network.  This column contains a unique identifier for the
    # network’s registered country as specified by GeoNames.  This ID can
    # be used to look up the location information in the Location file.
    FIELD_NAME_REGISTERED_COUNTRY_GEONAME_ID = 'registered_country_geoname_id'

    # The represented country is the country which is represented by users
    # of the IP address. For instance, the country represented by an
    # overseas military base. This column contains a unique identifier for
    # the network’s registered country as specified by GeoNames.  This ID
    # can be used to look up the location information in the Location file.
    FIELD_NAME_REPRESENTED_COUNTRY_GEONAME_ID = 'represented_country_geoname_id'

    # Deprecated. Please see our GeoIP2 Anonymous IP database
    # (https://www.maxmind.com/en/geoip2-anonymous-ip-database) to
    # determine whether the IP address is used by an anonymizing service.
    FIELD_NAME_IS_ANONYMOUS_PROXY = 'is_anonymous_proxy'

    # Deprecated. Please see our GeoIP2 Anonymous IP database.
    FIELD_NAME_IS_SATELLITE_PROVIDER = 'is_satellite_provider'

    # The postal code associated with the IP address.  These are available
    # for some IP addresses in Australia, Canada, France, Germany, Italy,
    # Spain, Switzerland, United Kingdom, and the US.  We return the first
    # 3 characters for Canadian postal codes.  We return the the first 2-4
    # characters (outward code) for postal codes in the United Kingdom.
    FIELD_NAME_POSTAL_CODE = 'postal_code'

    # The approximate latitude of the location associated with the network.
    # This value is not precise and should not be used to identify a
    # particular address or household.
    FIELD_NAME_LATITUDE = 'latitude'

    # The approximate longitude of the location associated with the network.
    # Latitude and Longitude are often near the center of population.  These
    # values are not precise and should not be used to identify a particular
    # address or household.
    FIELD_NAME_LONGITUDE = 'longitude'

    # The radius in kilometers around the specified location where the IP
    # address is likely to be.
    FIELD_NAME_ACCURACY_RADIUS = 'accuracy_radius'

    FIELD_NAMES = [
        FIELD_NAME_NETWORK,
        FIELD_NAME_GEONAME_ID,
        FIELD_NAME_REGISTERED_COUNTRY_GEONAME_ID,
        FIELD_NAME_REPRESENTED_COUNTRY_GEONAME_ID,
        FIELD_NAME_IS_ANONYMOUS_PROXY,
        FIELD_NAME_IS_SATELLITE_PROVIDER,
        FIELD_NAME_POSTAL_CODE,
        FIELD_NAME_LATITUDE,
        FIELD_NAME_LONGITUDE,
        FIELD_NAME_ACCURACY_RADIUS
    ]


    COLUMN_NAME_NETWORK = 'network'
    COLUMN_NAME_GEONAME_ID = 'geoname_id'
    COLUMN_NAME_REGISTERED_COUNTRY_GEONAME_ID = 'registered_country_geoname_id'
    COLUMN_NAME_REPRESENTED_COUNTRY_GEONAME_ID = 'represented_country_geoname_id'
    COLUMN_NAME_POSTAL_CODE = 'postal_code'
    COLUMN_NAME_LOCATION = 'location'
    COLUMN_NAME_ACCURACY = 'accuracy'

    COLUMN_NAMES = [
        COLUMN_NAME_NETWORK,
        COLUMN_NAME_GEONAME_ID,
        COLUMN_NAME_REGISTERED_COUNTRY_GEONAME_ID,
        COLUMN_NAME_REPRESENTED_COUNTRY_GEONAME_ID,
        COLUMN_NAME_POSTAL_CODE,
        COLUMN_NAME_LOCATION,
        COLUMN_NAME_ACCURACY
    ]

    @staticmethod
    def csv_to_sql(fd):
        """
        Generate the SQL values representing each block contained in the
        GeoIP2 block CSV file.


        @warning: the function makes the assumption that the CSV fields are
            given in the order described in MaxMind's specifications, i.e., as
            declared in the array ``FIELD_NAMES``.


        @param fd: file descriptor of a GeoIP2 block CSV file.


        @return: a generator of SQL values representing each block contained
            in the GeoIP2 block CSV file.
        """
        # Retrieve the names of the fields from the GeoIP2 block CSV file.
        field_names = fd.readline().strip().split(',')

        # Check whether some fields are unknown or missing.
        unknown_field_names = [ field_name for field_name in field_names
                if field_name not in GeoIP2BlockFile.FIELD_NAMES ]
        assert not unknown_field_names, 'Unsupported GeoIP2 block field(s): ' % ', '.join(unknown_field_names)

        missing_field_names = [ field_name for field_name in GeoIP2BlockFile.FIELD_NAMES
                if field_name not in field_names ]
        assert not missing_field_names, 'Missing GeoIP2 block field(s): ' % ', '.join(missing_field_names)

        # Parse each line of GeoIP2 block and generate the corresponding SQL
        # values.
        for line in fd:
            (network,
             geoname_id,
             registered_country_geoname_id,
             represented_country_geoname_id,
             _, # @deprecated is_anonymous_proxy
             _, # @deprecated is_satellite_provider
             postal_code,
             latitude,
             longitude,
             radius) = line.strip().split(',')

            yield [
                network,
                geoname_id,
                registered_country_geoname_id,
                represented_country_geoname_id,
                postal_code,
                longitude and latitude and 'SRID=4326;POINT(%s %s)' % (longitude, latitude),
                radius ]


class GeoIP2LocationFile(object):
    # Regular expression that matches the name of a GeoIP2 location CSV
    # file.
    REGEX_GEOIP_LOCATION_CSV_FILE_NAME = re.compile(r'GeoLite2-City-Locations-(([a-z]{2,3}-[A-Z]{2})|[a-z]{2,3}).csv')

    # Regular expression that matches a value in a CSV file that is not
    # properly quoted.
    REGEX_QUOTED_UNQUOTED_CSV_VALUE = re.compile(r'(?:^|,)(\"(?:[^\"]+|\"\")*\"|[^,]*)')

    # A unique identifier for the a location as specified by GeoNames.  This
    # ID can be used as a key for the Location file.
    FIELD_NAME_GEONAME_ID = 'geoname_id'

    # The locale that the names in this row are in. This will always
    # correspond to the locale name of the file.
    FIELD_NAME_LOCALE_CODE = 'locale_code'

    # The continent code for this location. Possible codes are:
    #
    #   * ``AF``: Africa
    #   * ``AN``: Antarctica
    #   * ``AS``: Asia
    #   * ``EU``: Europe
    #   * ``NA``: North America
    #   * ``OC``: Oceania
    #   * ``SA``: South America
    FIELD_NAME_CONTINENT_CODE = 'continent_code'

    # The continent name for this location in the file’s locale.
    FIELD_NAME_CONTINENT_NAME = 'continent_name'

    # A two-character ISO 3166-1 country code for the country associated
    # with the location.
    FIELD_NAME_COUNTRY_ISO_CODE = 'country_iso_code'

    # The country name for this location in the file’s locale.
    FIELD_NAME_COUNTRY_NAME = 'country_name'

    # A string of up to three characters containing the region-portion of
    # the ISO 3166-2 code for the first level region associated with the IP
    # address. Some countries have two levels of subdivisions, in which case
    # this is the least specific. For example, in the United Kingdom this
    # will be a country like “England”, not a county like “Devon”.
    FIELD_NAME_SUBDIVISION_1_ISO_CODE = 'subdivision_1_iso_code'

    # The subdivision name for this location in the file’s locale.  As with
    # the subdivision code, this is the least specific subdivision for the
    # location.
    FIELD_NAME_SUBDIVISION_1_NAME = 'subdivision_1_name'

    # A string of up to three characters containing the region-portion of
    # the ISO 3166-2 code for the second level region associated with the
    # IP address.  Some countries have two levels of subdivisions, in which
    # case this is the most specific. For example, in the United Kingdom
    # this will be a a county like “Devon”, not a country like “England”.
    FIELD_NAME_SUBDIVISION_2_ISO_CODE = 'subdivision_2_iso_code'

    # The subdivision name for this location in the file’s locale.  As with
    # the subdivision code, this is the most specific subdivision for the
    # location.
    FIELD_NAME_SUBDIVISION_2_NAME = 'subdivision_2_name'

    # The city name for this location in the file’s locale.
    FIELD_NAME_CITY_NAME = 'city_name'

    # The metro code associated with the IP address.  These are only
    # available for networks in the US. MaxMind provides the same metro
    # codes as used by DoubleClick.
    FIELD_NAME_METRO_CODE = 'metro_code'

    # The time zone associated with location, as specified by the IANA Time
    # Zone Database, e.g., “America/New_York”.
    FIELD_NAME_TIME_ZONE = 'time_zone'
    
    FIELD_NAMES = [
        FIELD_NAME_GEONAME_ID,
        FIELD_NAME_LOCALE_CODE,
        FIELD_NAME_CONTINENT_CODE,
        FIELD_NAME_CONTINENT_NAME,
        FIELD_NAME_COUNTRY_ISO_CODE,
        FIELD_NAME_COUNTRY_NAME,
        FIELD_NAME_SUBDIVISION_1_ISO_CODE,
        FIELD_NAME_SUBDIVISION_1_NAME,
        FIELD_NAME_SUBDIVISION_2_ISO_CODE,
        FIELD_NAME_SUBDIVISION_2_NAME,
        FIELD_NAME_CITY_NAME,
        FIELD_NAME_METRO_CODE,
        FIELD_NAME_TIME_ZONE
    ]

    @staticmethod
    def csv_to_sql(fd):
        # Retrieve the names of the fields from the GeoIP2 location CSV file.
        field_names = fd.readline().strip().split(',')

        # Check whether some fields are unknown or missing.
        unknown_field_names = [ field_name for field_name in field_names
                if field_name not in GeoIP2LocationFile.FIELD_NAMES ]
        assert not unknown_field_names, 'Unsupported GeoIP2 block field(s): ' % ', '.join(unknown_field_names)

        missing_field_names = [ field_name for field_name in GeoIP2LocationFile.FIELD_NAMES
                if field_name not in field_names ]
        assert not missing_field_names, 'Missing GeoIP2 block field(s): ' % ', '.join(missing_field_names)

        # Parse each line of GeoIP2 block and generate the corresponding SQL
        # values.
        for line in fd:
            values = GeoIP2LocationFile.REGEX_QUOTED_UNQUOTED_CSV_VALUE.findall(line.strip())
            assert len(values) > 0, 'The specified entry does not correspond to a valid CSV line'
            yield [ value.replace('"', '') for value in values ]


def fetch_latest_geolite2_zip_file(url, path_name):
    """
    Download the ZIP archive of the latest version of the GeoLite2
    database.

    The function checks whether a previous ZIP archive of the GeoLite2
    database has been downloaded, and whether this archive corresponds to
    the latest version available.  If so, the function immediately returns
    the path and name of this file.


    @param url: Uniform Resource Locator (URL) of the ZIP archive of the
        latest version of the GeoLite2 database hosted on MaxMind's server.

    @param path_name: name of the local path where to store this ZIP
        archive file.


    @return: path and name of the ZIP archive file of the latest version
        of the GeoLite2 database is stored in.
    """
    file_util.make_directory_if_not_exists(path_name)

    zip_file_name = os.path.basename(url)
    geolite2_zip_file_path_name = os.path.join(path_name, zip_file_name)
    geolite2_md5_zip_file_path_name = os.path.join(path_name, zip_file_name + '.md5')

    opener = urllib2.build_opener()
    opener.addheaders = [ ('User-agent', 'Mozilla/5.0') ]

    with contextlib.closing(opener.open(GEOLITE2_CITY_ZIP_URL + '.md5', None, 5)) as url_handle:
        md5_digest = url_handle.read()

    if os.path.exists(geolite2_md5_zip_file_path_name):
        with open(geolite2_md5_zip_file_path_name, 'rt') as file_handle:
            previous_md5_digest = file_handle.read()
            if md5_digest == previous_md5_digest:
                return geolite2_zip_file_path_name

    with contextlib.closing(opener.open(url, None, 5)) as url_handle:
        with open(geolite2_zip_file_path_name, 'wb') as file_handle:
            file_handle.write(url_handle.read())

    with open(geolite2_md5_zip_file_path_name, 'wt') as file_handle:
        file_handle.write(md5_digest)

    return geolite2_zip_file_path_name


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--path', dest='path_name', metavar='pathname', default='~/.geolite2',
            help='specify the absolute path of the directory GeoLite ZIP archive is stored in')
    arguments = parser.parse_args()

    # Fetch the ZIP archive file of the latest version of the GeoLite2
    # database, whether from the cache, whether from MaxMind's server.
    zip_file_path_name = fetch_latest_geolite2_zip_file(GEOLITE2_CITY_ZIP_URL, os.path.expanduser(arguments.path_name))
    zip_file = zipfile.ZipFile(zip_file_path_name)

    # Parse the GeoIP2 block CSV file stored as an entry of the ZIP archive.
    for entry_file_name in zip_file.filelist:
        if os.path.basename(entry_file_name.filename) == GeoIP2BlockFile.GEOIP_BLOCK_CSV_FILE_NAME:
            print 'COPY geoip_block(%s) FROM stdin;' % ','.join(GeoIP2BlockFile.COLUMN_NAMES)

            fd = zip_util.open_entry_file(zip_file, entry_file_name)
            for values in GeoIP2BlockFile.csv_to_sql(fd):
                print '\t'.join([ value or r'\N' for value in values ])

            print '\\.\n'
            fd.close()

            break

    # Parse the various translated GeoIP location CSV files stored as
    # entries of the ZIP archive.
    geoip2_location_file_exists = False

    for entry_file_name in zip_file.filelist:
        if GeoIP2LocationFile.REGEX_GEOIP_LOCATION_CSV_FILE_NAME.match(os.path.basename(entry_file_name.filename)):
            if not geoip2_location_file_exists:
                print 'COPY geoip_location(%s) FROM stdin;' % ','.join(GeoIP2LocationFile.FIELD_NAMES)
                geoip2_location_file_exists = True

            fd = zip_util.open_entry_file(zip_file, entry_file_name)
            for values in GeoIP2LocationFile.csv_to_sql(fd):
                print '\t'.join([ value or r'\N' for value in values ])

            fd.close()

        if geoip2_location_file_exists:
            geoip2_location_file_exists = False
            print '\\.\n'



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import json

from majormode.perseus.constant.obj import ObjectStatus
from majormode.perseus.model.obj import Object
from majormode.perseus.service.base_rdbms_service import BaseRdbmsService
from majormode.perseus.service.base_rdbms_service import BaseService
from majormode.perseus.utils import cast
from majormode.perseus.utils.date_util import ISO8601DateTime
from pymemcache.client import base

import settings


class SessionService(BaseRdbmsService):
    """
    A login session is the period of activity between a user sign-in
    and sign-out of the platform.
    """

    class ExpiredSessionException(BaseService.BaseServiceException):
        """
        Signal that the account session of the user has expired.  The user
        needs to login again.
        """
        pass

    # Default user login session duration (token lifetime) expressed in
    # minutes (90 days).
    DEFAULT_LOGIN_SESSION_DURATION = 60 * 24 * 90

    # Maximum user login session duration (token lifetime) expressed in
    # minutes (180 days).
    MAXIMUM_LOGIN_SESSION_DURATION = 60 * 24 * 180

    # Pattern of the key used to store user login session providing the
    # identification of this session.
    MEMCACHED_SESSION_KEY_PATTERN = 'session:login:%s'

    def __init__(
            self,
            memcached_server_hostname=settings.MEMCACHED_SERVER_HOSTNAME or '127.0.0.1',
            memcached_server_port_number=settings.MEMCACHED_SERVER_PORT or 11211):
        """
        Build a new instance of the session service


        :param memcached_server_hostname: host name of the memcached server
            used to store user login session.

        :param memcached_server_port_number: port number of the memcached
            server.
        """
        super(SessionService, self).__init__()
        self._memcached = base.Client((memcached_server_hostname, memcached_server_port_number))

    def create_session(
            self,
            app_id,
            account_id,
            connection=None,
            session_duration=None):
        """
        Create a login session for the specified user.  The identification of
        the session can be used as an authentication token.


        :param app_id: Identification of the client application such as a Web,
            a desktop, or a mobile application, that accesses the service.

        :param account_id: Identification of a user account.

        :param connection: An `RdbmsConnection` object supporting the Python

        :param session_duration: Login session duration, expressed in
            seconds, corresponding to the interval of time between the
            creation of the token and the expiration of this login session.


        :return: an object containing the following members:

            * ``session_id``: identification of the user login session, also known
              as the authentication token.

            * ``account_id``: identification of the account of the user.

            * ``creation_time``: time when this login session has been created.

            * ``expiration_time``: time when this login session is going to expire.
        """
        session_duration = min(
            session_duration or SessionService.DEFAULT_LOGIN_SESSION_DURATION,
            SessionService.MAXIMUM_LOGIN_SESSION_DURATION)

        # Check whether a login session already exists for the specified user
        # and API key.  If any, reuse this login session by extending its
        # expiration time by the given duration, otherwise create a new login
        # session for this user with this API key.
        with self.acquire_rdbms_connection(connection=connection, auto_commit=True) as connection:
            cursor = connection.execute(
                """
                UPDATE 
                    account_session
                  SET 
                    object_status = %(OBJECT_STATUS_ENABLED)s,
                    expiration_time = current_timestamp + '%(session_duration)s minutes'::interval,
                    update_time = current_timestamp
                  FROM (
                    SELECT 
                        session_id
                      FROM
                        account_session
                      WHERE
                        account_id = %(account_id)s
                        AND app_id = %(app_id)s) AS self
                  WHERE
                    account_session.session_id = self.session_id
                  RETURNING
                    account_session.session_id,
                    account_id,
                    app_id,
                    creation_time,
                    expiration_time
                """,
                {
                    'OBJECT_STATUS_ENABLED': ObjectStatus.enabled,
                    'account_id': account_id,
                    'app_id': app_id,
                    'session_duration': session_duration
                })
            row = cursor.fetch_one()

            if row is None:
                cursor = connection.execute(
                    """
                    INSERT INTO account_session(
                        account_id,
                        app_id,
                        expiration_time)
                      VALUES (
                        %(account_id)s,
                        %(app_id)s,
                        current_timestamp + '%(session_duration)s minutes'::interval)
                      RETURNING 
                        session_id,
                        account_id,
                        creation_time,
                        app_id,
                        expiration_time""",
                    {
                        'account_id': account_id,
                        'app_id': app_id,
                        'session_duration': session_duration
                    })
                row = cursor.fetch_one()

            session = row.get_object({
                'account_id': cast.string_to_uuid,
                'app_id': cast.string_to_uuid,
                'creation_time': cast.string_to_timestamp,
                'expiration_time': cast.string_to_timestamp,
                'session_id': cast.string_to_uuid
            })

        # Set the session token into the cache for future references.
        self._memcached.set(
            SessionService.MEMCACHED_SESSION_KEY_PATTERN % session.session_id.hex,
            json.dumps(session.stringify()),
            expire=session_duration)

        return session

    def drop_session(
            self,
            app_id,
            session,
            connection=None):
        """
        Drop the specified login session of a user.


        :param app_id: identification of the client application such as a Web,
            a desktop, or a mobile application, that accesses the service.

        :param session: An object containing the following attributes:

            * `account_id` (required): Identification of the account of the a user.

            * `session_id` (required): Identification of the user's session.

        :param connection: An object `RdbmsConnection`.


        :return: an object containing the following attribute:

            * ``session_id``: identification of the session that has been deleted.


        :raise IllegalAccessException: if the specified login session doesn't
            belong to the specified user.

        :raise UndefinedObjectException: if the specified identification
            doesn't refer to any user login session registered to the platform.
        """
        with self.acquire_rdbms_connection(auto_commit=True, connection=connection) as connection:
            cursor = connection.execute(
                """
                UPDATE 
                    account_session
                  SET 
                    object_status = %(OBJECT_STATUS_DELETED)s
                  WHERE 
                    account_id = %(account_id)s
                    AND app_id = %(app_id)s
                    AND object_status <> %(OBJECT_STATUS_DELETED)s
                  RETURNING 
                    session_id
                """,
                {
                    'OBJECT_STATUS_DELETED': ObjectStatus.deleted,
                    'account_id': session.account_id,
                    'app_id': app_id
                })

            row = cursor.fetch_one()
            if row is None:
                raise SessionService.UndefinedObjectException(
                        "undefined user session",
                        payload={'account_id': session.account_id, 'app_id': app_id })

            session = row.get_object({'session_id': cast.string_to_uuid})

            if session.session_id != session.session_id:
                raise SessionService.IllegalAccessException(
                    "the session does not belong to the specified user account",
                    payload={
                        'account_id': session.account_id,
                        'session_id': session.session_id
                    })

        self._memcached.delete(SessionService.MEMCACHED_SESSION_KEY_PATTERN % session.session_id.hex)

        return session

    def get_session(self, app_id, session_id):
        """
        Return the information about the specified authentication session.


        :param app_id: identification of the client application such as a Web,
            a desktop, or a mobile application, that accesses the service.

        :param session_id: identification of a user session.


        :return: a ``SessionToken`` instance containing the following members:

            * ``session_id``: authentication token that references a login session
              of a user.

            * ``account_id``: identification of the account of the user who has
                signed-in.

            * ``app_id``: identification of the client application such as a Web,
              a desktop, or a mobile application, that accesses the service.

            * ``creation_time``: date and time when the authentication token has
              been created, i.e., when the user has signed-in.

            * ``expiration_time``: date and time when the authentication token is
              going to expire.


        :raise ExpiredSessionException: If the account session of the user has
            expired.  The user needs to login again.


        :raise IllegalAccessException: If the user login session is registered
            with an API key other than the one specified.

        :raise UndefinedObjectException: Indicate that the specified
            identification doesn't refer to any user login session registered
            to the platform.
        """
        # Try first to retrieve the session from the cache before trying to
        # retrieve it from database.
        payload = self._memcached.get(SessionService.MEMCACHED_SESSION_KEY_PATTERN % session_id.hex)
        if False:  # deactivate cache
            # if payload is not None:
            session = Object()
            session.__dict__.update(json.loads(payload))
            session.account_id = cast.string_to_uuid(session.account_id)
            session.app_id = cast.string_to_uuid(session.app_id)
            session.creation_id = cast.string_to_timestamp(session.creation_time)
            session.expiration_time = cast.string_to_timestamp(session.expiration_time)
            session.session_id = cast.string_to_uuid(session.session_id)

            if session.expiration_time < ISO8601DateTime.now():
                raise self.ExpiredSessionException('The user session has timeout')

        else:
            with self.acquire_rdbms_connection() as connection:
                cursor = connection.execute(
                    """
                    SELECT 
                        session_id,
                        account_id,
                        app_id,
                        creation_time,
                        expiration_time,
                        object_status
                      FROM
                        account_session
                      WHERE
                        session_id = %(session_id)s
                    """,
                    {
                        'session_id': session_id
                    })
                row = cursor.fetch_one()
                if row is None:
                    raise self.ExpiredSessionException('Undefined user session')

                session = row.get_object({
                    'account_id': cast.string_to_uuid,
                    'app_id': cast.string_to_uuid,
                    'creation_time': cast.string_to_timestamp,
                    'expiration_time': cast.string_to_timestamp,
                    'session_id': cast.string_to_uuid
                })

                # Check the current status of the user session.
                if session.object_status == ObjectStatus.disabled:
                    raise self.ExpiredSessionException('The user has been logout')
                elif session.object_status == ObjectStatus.deleted:
                    raise self.ExpiredSessionException('The user has logout')

                del session.object_status

                # Check whether the user session has expired.
                if session.expiration_time < ISO8601DateTime.now():
                    raise self.ExpiredSessionException('The user session has timeout')

            # Store the user session into the cache for future references.
            self._memcached.set(
                    SessionService.MEMCACHED_SESSION_KEY_PATTERN % session_id.hex,
                    json.dumps(session.stringify()),
                    expire=SessionService.DEFAULT_LOGIN_SESSION_DURATION)

        if session.app_id != app_id:
            raise SessionService.IllegalAccessException(
                'The session is registered with another application',
                payload={'session_id': session_id})

        return session



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from majormode.perseus.constant.http import HttpMethod
from majormode.perseus.model.geolocation import GeoPoint
from majormode.perseus.model.locale import DEFAULT_LOCALE
from majormode.perseus.service.base_http_handler import HttpRequest
from majormode.perseus.service.base_http_handler import HttpRequestHandler
from majormode.perseus.service.base_http_handler import http_request
from majormode.perseus.service.area.area_service import AreaService


class AreaServiceHttpRequestHandler(HttpRequestHandler):
    pass
    # @http_request(r'^/area$',
    #               http_method=HttpMethod.GET,
    #               authentication_required=False)
    # def get_areas(self, request):
    #     locale = request.get_argument('locale',
    #             data_type=HttpRequest.ArgumentDataType.locale,
    #             default_value=DEFAULT_LOCALE)
    #
    #     # Return areas specified by their identification.
    #     area_id_list = request.get_argument('ids',
    #             data_type=HttpRequest.ArgumentDataType.list,
    #             item_data_type=HttpRequest.ArgumentDataType.uuid,
    #             is_required=False)
    #
    #     if area_id_list:
    #         return AreaService().get_areas(request.app_id, area_id_list, locale=locale)
    #
    #     highest_area_level = request.get_argument('highest_area_level',
    #             data_type=HttpRequest.ArgumentDataType.integer,
    #             is_required=False)
    #
    #     lowest_area_level = request.get_argument('lowest_area_level',
    #             data_type=HttpRequest.ArgumentDataType.integer,
    #             is_required=False)
    #
    #     minimal_area_surface = request.get_argument(
    #         'minimal_area_surface',
    #         data_type=HttpRequest.ArgumentDataType.decimal,
    #         is_required=False)
    #
    #     # Return the areas that encompass the specified location.
    #     location_url_value = request.get_argument(
    #         'll',
    #         data_type=HttpRequest.ArgumentDataType.string,
    #         is_required=False)
    #
    #     if location_url_value:
    #         location_coordinates = location_url_value.split(',')
    #         if len(location_coordinates) != 2:
    #             raise AreaService.InvalidArgumentException('The location MUST have the form "lon,lat"')
    #
    #         location = GeoPoint(float(location_coordinates[0]), float(location_coordinates[1]))
    #
    #         return AreaService().get_areas_by_location(
    #             request.app_id, location,
    #             highest_area_level=highest_area_level,
    #             locale=locale,
    #             lowest_area_level=lowest_area_level,
    #             minimal_area_surface=minimal_area_surface)
    #
    #     # Return the areas that encompass the approximate location corresponding
    #     # to the specified IPv4 address.
    #     ip_address = request.get_argument(
    #         'ip_address',
    #         data_type=HttpRequest.ArgumentDataType.ipv4,
    #         is_required=False)
    #
    #     if ip_address:
    #         return AreaService().get_areas_by_ip_address(
    #             request.app_id,
    #             ip_address,
    #             highest_area_level=highest_area_level,
    #             locale=locale,
    #             lowest_area_level=lowest_area_level,
    #             minimal_area_surface=minimal_area_surface)
    #
    #     # Return the areas that are including in the specified boundaries.
    #     bounds_url_value = request.get_argument(
    #         'bounds',
    #         data_type=HttpRequest.ArgumentDataType.string,
    #         is_required=False)
    #
    #     if bounds_url_value:
    #         bounds_coordinates = bounds_url_value.split(',')
    #         if len(bounds_coordinates) != 4:
    #             raise AreaService.InvalidArgumentException('The bounds MUST have the form "NE-lon,NE-lat,SW-lon,SW-lat"')
    #
    #         bounds = (GeoPoint(float(bounds_coordinates[0]), float(bounds_coordinates[1])),
    #                   GeoPoint(float(bounds_coordinates[2]), float(bounds_coordinates[3])))
    #
    #         return AreaService().get_areas_in_bounding_box(request.app_id, bounds, locale=locale)
    #
    #     # Return areas that match the specified keywords.
    #     keywords = request.get_argument(
    #         'keywords',
    #         data_type=HttpRequest.ArgumentDataType.list,
    #         is_required=True)
    #
    #     offset = request.get_argument(
    #         'offset',
    #         data_type=HttpRequest.ArgumentDataType.integer,
    #         is_required=False,
    #         default_value=0)
    #
    #     limit = request.get_argument(
    #         'limit',
    #         data_type=HttpRequest.ArgumentDataType.integer,
    #         is_required=False,
    #         default_value=AreaService.MAXIMUM_LIMIT)
    #
    #     return AreaService().get_areas_by_keywords(
    #         request.app_id,
    #         keywords,
    #         highest_area_level=highest_area_level,
    #         limit=limit,
    #         locale=locale,
    #         lowest_area_level=lowest_area_level,
    #         minimal_area_surface=minimal_area_surface,
    #         offset=offset)
    #
    # @http_request(r'^/area/(area_id:uuid)/boundaries$',
    #               http_method=HttpMethod.GET,
    #               authentication_required=False)
    # def get_area_boundaries(self, request, area_id):
    #     return AreaService().get_area_boundaries(request.app_id, area_id)



# Copyright (C) 2019 Majormode.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from majormode.perseus.service.team.team_service import TeamService

import settings
import smtplib
import time

# Define the minimum time interval between sending batch of invites,
# in seconds.
MINIMUM_TIME_INTERVAL_BETWEEN_INVITE_SENDING = 10

if __name__ == "__main__":
    pass
    # print "Team Member Invite Email Sender"
    # print "Copyright (C) 2013 Majormode.  All rights reserved."
    #
    # while True:
    #     try:
    #         invites = TeamService().send_invites(settings.PLATFORM_BOTNET_APP_ID,
    #                 settings.PLATFORM_BOTNET_ACCOUNT_ID)
    #
    #         print '[INFO] The following users have been sent an invite email:'
    #         for invite in invites:
    #             print '\t• %(recipient_name)s (%(email_address)s) invited to team %(team_name)s' % \
    #                 { 'email_address': invite.email_address,
    #                   'recipient_name': invite.recipient_name,
    #                   'team_name': invite.team_name }
    #
    #         time.sleep(MINIMUM_TIME_INTERVAL_BETWEEN_INVITE_SENDING)
    #     except smtplib.SMTPDataError, error:
    #         print '[ERROR] %s; waiting for one day...' % error.smtp_error
    #         time.sleep(60 * 60 * 24)
    #     except:
    #         print '[ERROR] $s; waiting for one minute...' % error.smtp_error
    #         time.sleep(60)
